#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
"""
Provides the StatusManager class.
"""

import typing
import operator
import abc
from PySide2.QtCore import QObject
from functools import wraps

if typing.TYPE_CHECKING:
    Signal = typing.Any
else:
    from PySide2.QtCore import Signal

T = typing.TypeVar("T")
S = typing.TypeVar("S")


class Status(typing.Generic[T]):
    """
    Manages a single value and notifies when this value is changed.
    """

    def __init__(self, value_type: typing.Type[T]) -> None:
        self.__communicator = self.create_communicator(value_type)

    @abc.abstractmethod
    def get(self) -> T:
        """
        Returns the current status.
        """

    class BaseCommunicator(QObject):

        def __init__(self, signal_type: typing.Type[T]) -> None:
            super().__init__()
            self.signal_type = signal_type

        def emit_signal(self, value: T) -> None:
            if not isinstance(value, self.signal_type):
                raise TypeError(f"Expected {self.signal_type}, got {type(value).__name__}")
            self.signal.emit(value)

    def create_communicator(self, signal_type: typing.Type[T]):
        class DynamicCommunicator(self.BaseCommunicator):  # type: ignore
            signal = Signal(signal_type)

            def __init__(self) -> None:
                super().__init__(signal_type)

        return DynamicCommunicator()

    @property
    def changed_signal(self) -> Signal:
        """
        Returns the signal used to signal changes.
        """
        return self.__communicator.signal

    @property
    def value(self) -> T:
        """
        Returns the current status.
        """
        return self.get()


class WriteableStatus(Status[T]):
    """
    Manages a single value and notifies when this value is changed.
    """

    @abc.abstractmethod
    def set(self, value: T) -> None:
        """
        Returns the current status.
        """

    @property
    def value(self) -> T:
        """
        Returns the current status.
        """
        return self.get()

    @value.setter
    def value(self, value: T) -> None:
        """
        Sets the current status.
        """
        self.set(value)


class StatusManager(WriteableStatus[T]):
    """
    Stores and manages a single value and notifies when this value is changed.
    """

    def __init__(self, value: T):
        super().__init__(type(value))
        self.__value = value

    def get(self) -> T:
        """
        Returns the contained value.
        """
        return self.__value

    def set(self, value: T) -> None:
        """
        Sets the contained value. Notifies on change.
        """
        if value == self.__value:
            return
        self.__value = value
        self.changed_signal.emit(self.__value)


class TransformedStatusManager(Status[T]):
    """
    Manages a value generated by a transform applied to Status instances.
    """

    def __init__(
        self,
        transform: typing.Callable[..., T],
        *args: Status[typing.Any],
        **kwargs: Status[typing.Any],
    ):
        self.__apply_transform: typing.Callable[[], T] = lambda: transform(
            *[arg.value for arg in args],
            **{key: value.value for key, value in kwargs.items()},
        )
        super().__init__(type(self.__apply_transform()))
        self.__value = StatusManager(self.__apply_transform())
        self.__value.changed_signal.connect(
            lambda: self.changed_signal.emit(self.__value.value)
        )
        for arg in args:
            arg.changed_signal.connect(self.__update)
        for _, value in kwargs.items():
            value.changed_signal.connect(self.__update)

    def get(self) -> T:
        """
        Returns the contained value.
        """
        return self.__value.value

    def __update(self) -> None:
        """
        Updates the contained value.
        """
        setattr(self.__value, "value", self.__apply_transform())


def lift(transform: typing.Callable[..., T]) -> typing.Callable[..., Status[T]]:
    """
    Converts a function `transform` to a function on Status instances.
    """

    @wraps(transform)
    def transformed(
        *args: typing.Any, **kwargs: typing.Any
    ) -> TransformedStatusManager[T]:
        return TransformedStatusManager[T](transform, *args, **kwargs)

    return transformed


and_ = lift(operator.and_)
or_ = lift(operator.or_)


class TransformedWriteableStatusManager(typing.Generic[S, T], WriteableStatus[T]):
    """
    Manages a writeable value generated by a transform applied to Status instances.
    """

    def __init__(
        self,
        transform: typing.Callable[[S], T],
        inverse: typing.Callable[[T], S],
        argument: WriteableStatus[S],
    ):
        self.__apply_transform: typing.Callable[[], T] = lambda: transform(
            argument.value
        )
        super().__init__(type(self.__apply_transform()))
        self.__apply_inverse: typing.Callable[[T], None] = lambda s: setattr(
            argument, "value", inverse(s)
        )
        self.__value = StatusManager(self.__apply_transform())
        self.__value.changed_signal.connect(
            lambda: self.changed_signal.emit(self.__value.value)
        )
        argument.changed_signal.connect(self.__update)

    def get(self) -> T:
        """
        Returns the contained value.
        """
        return self.__value.value

    def set(self, value: T) -> None:
        """
        Sets the contained value.
        """
        self.__apply_inverse(value)

    def __update(self) -> None:
        """
        Updates the contained value.
        """
        setattr(self.__value, "value", self.__apply_transform())


def lift_writeable(
    transform: typing.Callable[[S], T], inverse: typing.Callable[[T], S]
) -> typing.Callable[[WriteableStatus[S]], WriteableStatus[T]]:
    """
    Converts a pair function `transform`, `inverse` to a function on
    WriteableStatus instances.
    """

    def transformed(
        argument: WriteableStatus[S],
    ) -> TransformedWriteableStatusManager[S, T]:
        return TransformedWriteableStatusManager[S, T](transform, inverse, argument)

    return transformed


not_ = lift_writeable(operator.not_, operator.not_)


class AdaptedStatusManager(Status[T]):
    """
    Manages a value generated using the provided functions.
    """

    def __init__(
        self, pull: typing.Callable[[], T], signal: Signal,
    ):
        super().__init__(type(pull()))
        self.__pull = pull

        signal.connect(lambda: self.changed_signal.emit(pull()))

    def get(self) -> T:
        """
        Returns the value returned by pull.
        """
        return self.__pull()


def as_status(pull: typing.Callable[[], T], signal: Signal) -> AdaptedStatusManager[T]:
    """
    Adapts a triple of functions to a Status.
    """
    return AdaptedStatusManager[T](pull=pull, signal=signal)


class AdaptedWriteableStatusManager(WriteableStatus[T]):
    """
    Manages a writeable value generated using the provided functions.
    """

    def __init__(
        self,
        pull: typing.Callable[[], T],
        push: typing.Callable[[T], None],
        signal: Signal,
    ):
        super().__init__(type(pull()))
        self.__pull = pull
        self.__push = push

        signal.connect(lambda: self.changed_signal.emit(pull()))

    def get(self) -> T:
        """
        Returns the value returned by pull.
        """
        return self.__pull()

    def set(self, value: T) -> None:
        """
        Sets the contained value using push.
        """
        self.__push(value)


def as_writeable_status(
    pull: typing.Callable[[], T], push: typing.Callable[[T], None], signal: Signal
) -> AdaptedWriteableStatusManager[T]:
    """
    Adapts a triple of functions to a WriteableStatus.
    """
    return AdaptedWriteableStatusManager[T](pull=pull, push=push, signal=signal)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

from typing import Tuple

from scine_chemoton.utilities.model_combinations import ModelCombination
from scine_chemoton.utilities.uncertainties import ConstantUncertainty
from scine_chemoton.utilities.place_holder_model import construct_place_holder_model


class UncertaintyTuple:
    def __init__(self, model_combination: ModelCombination, free_energy_uncertainty: ConstantUncertainty,
                 activation_energy_uncertainty: ConstantUncertainty):
        """
        We cannot directly call isinstance(option, Tuple[ModelCombination, Uncertainty, Uncertainty]. Therefore,
        we use this dummy object to make the comparison.

        Parameters
        ----------
        model_combination : ModelCombination
            The model combination for which the uncertainties are valid.
        free_energy_uncertainty : Uncertainty
            The uncertainty object for the free energies.
        activation_energy_uncertainty : Uncertainty
            The uncertainty object for the activation energies.
        """
        self.model_combination = model_combination
        self.free_energy_uncertainty = free_energy_uncertainty
        self.activation_energy_uncertainty = activation_energy_uncertainty

    def get_tuple(self) -> Tuple[ModelCombination, ConstantUncertainty, ConstantUncertainty]:
        """
        Getter as a tuple.
        """
        return self.model_combination, self.free_energy_uncertainty, self.activation_energy_uncertainty

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, UncertaintyTuple):
            return False
        return (other.model_combination == self.model_combination
                and other.free_energy_uncertainty.get_uncertainty_bounds()
                == self.free_energy_uncertainty.get_uncertainty_bounds()
                and other.activation_energy_uncertainty.get_uncertainty_bounds()
                == self.activation_energy_uncertainty.get_uncertainty_bounds())

    @staticmethod
    def get_default():
        return UncertaintyTuple(ModelCombination(construct_place_holder_model()),
                                ConstantUncertainty((0.0, 0.0)),
                                ConstantUncertainty((0.0, 0.0)))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

import os

from PySide2.QtWidgets import QTreeWidget, QTreeWidgetItem, QWidget
from PySide2.QtGui import QBrush, QColor, QPixmap

from scine_heron.autocas.autocas_settings import AutocasSettings
from scine_heron.autocas.signal_handler import SignalHandler
from scine_heron.styling.delegates import CustomLightDelegate
import scine_heron.config as config


class FileTreeWidget(QTreeWidget):
    def __init__(
        self,
        parent: QWidget,
        signal_handler: SignalHandler,
        autocas_settings: AutocasSettings,
    ):
        super().__init__(parent)
        if config.MODE in config.LIGHT_MODES:
            self.setItemDelegate(CustomLightDelegate())
        self.signal_handler = signal_handler
        self.settings = autocas_settings

        self.main_dir = os.getcwd()
        self.setHeaderLabels([self.main_dir])
        # self.setSelectionMode(QAbstractItemView.ExtendedSelection)
        # self.setColumnCount(1)
        # self.setAlternatingRowColors(False)
        self.fillTree()
        # self.header().setSectionResizeMode(QHeaderView.ResizeToContents)
        # self.header().setStretchLastSection(False)
        # self.header().setSectionResizeMode(5, QHeaderView.Stretch)
        self.show()
        # pylint: disable-next=E1101
        self.itemClicked.connect(self.load_item)

        self.visible = True
        self.expand_ico = QPixmap(":teRightArrow.png")
        self.collapse_ico = QPixmap(":teLeftArrow.png")
        self.signal_handler.toggle_file_tree.connect(self.__toggle)

    def __toggle(self):
        if self.visible:
            self.setVisible(False)
            self.visible = False
        else:
            self.setVisible(True)
            self.visible = True

    def fillTree(self):
        def iterate(currentDir, currentItem):
            for f in os.listdir(currentDir):
                path = os.path.join(currentDir, f)
                if f.startswith("."):
                    pass
                else:
                    if os.path.isdir(path):
                        dirItem = QTreeWidgetItem(currentItem)
                        dirItem.setText(0, f)
                        dirItem.setText(1, path)
                        iterate(path, dirItem)
                    elif not os.path.isdir(path):
                        dirItem = QTreeWidgetItem(currentItem)
                        dirItem.setText(0, f)
                        dirItem.setText(1, path)
                        if dirItem.text(0).endswith(".xyz") or dirItem.text(0).endswith(".molden")\
                                or dirItem.text(0).endswith("groups.dat"):
                            pass
                        else:
                            dirItem.setForeground(0, QBrush(QColor("red")))
        iterate(self.main_dir, self)

    def load_item(self, item):
        if item.text(0).endswith(".xyz"):
            # print("Loading xyz-file")
            self.signal_handler.open_molecule_widget_signal.emit()
            self.signal_handler.load_xyz_file_signal.emit(item.text(1))
        elif item.text(0).endswith(".molden"):
            # print("Loading molden-file")
            self.signal_handler.open_molecule_widget_signal.emit()
            self.signal_handler.load_molden_file_signal.emit(item.text(1))
        elif item.text(0).endswith("groups.dat"):
            # print("Loading orbital groups-file")
            self.signal_handler.load_orbital_groups_file_signal.emit(item.text(1))
            # TODO



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
from time import sleep
from typing import TYPE_CHECKING, Any, Dict, List, Optional

from PySide2.QtCore import QThread, QObject
from PySide2.QtWidgets import QLineEdit

if TYPE_CHECKING:
    Signal = Any
else:
    from PySide2.QtCore import Signal

from scine_chemoton.steering_wheel import FailedSaveException, SteeringWheel
from scine_chemoton.steering_wheel.selections import (
    Selection,
)
from scine_chemoton.steering_wheel.datastructures import (
    SelectionResult,
    Status,
    ExplorationSchemeStep,
    ExplorationResult,
    RestartPartialExpansionInfo,
)

from scine_database import Credentials

from scine_heron.io.text_box import text_input_box
from scine_heron.utilities import construct_sound


def _input(question: str) -> str:
    return text_input_box(None, "Scine Heron", question, QLineEdit.Normal, "y")


class WheelThread(QThread):
    """
    This Thread is constructed and controlled by the Steering Tab.
    When constructed, it constructs the underlying SteeringWheel and when started
    it continuously queries the SteeringWheel for certain properties and holds them in
    memory.
    It communicates mainly with its signals, but is also a wrapper around
    many SteeringWheel methods.
    """

    # signals that this threads sends to main display for transfer wheel information
    status_signal = Signal(int, Status)
    run_info_signal = Signal(bool)
    selection_signal = Signal(str, SelectionResult)

    # technical reasons
    error_signal = Signal(str)  # error message cannot be displayed from thread

    def __init__(self, parent: QObject, credentials: Credentials):
        """
        Construct the Thread and the SteeringWheel underneath.

        Parameters
        ----------
        parent : QObject
            The parent widget
        credentials : Credentials
            The credentials of the database the Wheel should work on.
        """
        super().__init__(parent)
        self._parent = parent
        self._stop = False
        self._paused = False
        self._sleep_time = 1.0
        self._last_selection_index: int = -1
        self._status_report: Dict[str, Status] = {}

        self._wheel = SteeringWheel(credentials, [], callable_input=_input)
        self._sound_start = construct_sound("engine-0")
        self._sound_stop = construct_sound("break")

    def start_wheel(self, restart: bool = True) -> None:
        """
        Starts the underlying wheel, this forks a process.

        Parameters
        ----------
        restart : bool, optional
            If the wheel should be restarted, if not, the results are deleted, by default True
        """
        if not restart:
            self._wheel.delete_results()
        self._wheel.run(allow_restart=restart, ask_for_how_many_results=True)
        if getattr(self._parent, "sound_allowed", False):
            self._sound_stop.stop()
            self._sound_start.play()

    def stop_wheel(self) -> None:
        """
        Stop the wheel gracefully.
        Whether the wheel tries to save its process, depends on the wheel defaults
        """
        if self._wheel.is_running():
            self._wheel.stop()
            if getattr(self._parent, "sound_allowed", False):
                self._sound_start.stop()
                self._sound_stop.play()
        self._wheel.join()

    def wheel_is_running(self) -> bool:
        """
        Whether the underlying wheel is currently running

        Returns
        -------
        bool
            The result
        """
        return self._wheel.is_running()

    def terminate_wheel(self, save_progress: bool) -> None:
        """
        Stop the wheel without waiting for the current step.
        """
        if getattr(self.parent(), "sound_allowed", False):
            self._sound_start.stop()
            self._sound_stop.play()
        self._wheel.terminate(try_save_progress=save_progress, suppress_warning=True)

    def terminate(self) -> None:
        """
        Stop the wheel without trying to save its progress.
        """
        if self.wheel_is_running():
            self.terminate_wheel(save_progress=False)
        super().terminate()

    def stop(self) -> None:
        """
        Stop the underlying wheel gracefully and also stop this thread.
        """
        self.stop_wheel()
        self._stop = True

    def save(self) -> None:
        """
        Save the current progress.
        """
        try:
            self._wheel.save()
        except FailedSaveException as e:
            self.error_signal.emit(f"Could not save {self._wheel.name} because {e}")

    def pause(self) -> None:
        """
        Pauses the constant result querying on the wheel by this thread.
        """
        self._paused = True

    def resume(self) -> None:
        """
        Resumes the constant result querying on the wheel by this thread.
        """
        self._paused = False

    def run(self) -> None:
        """
        Start the constant result querying on the wheel by this thread.
        """
        while not self._stop:
            while self._paused:
                sleep(0.1)
            try:
                self._run_info_check()
                self._status_check()
                self._selection_check()
            except BaseException as e:
                print(e)
                import traceback
                traceback.print_exc()
                self._abort_run()
            else:
                sleep(self._sleep_time)
        self.exit(0)

    def reset_selection_index(self) -> None:
        """
        Reset the thread's information what the latest active selection is.
        """
        self._last_selection_index = -1

    def _abort_run(self) -> None:
        """
        Stop the wheel and prints an error message
        """
        self.error_signal.emit(f"Had to stop {self._wheel.name} because an unhandable error occurred")
        self.stop_wheel()

    def _run_info_check(self) -> None:
        """
        Send the current run status as a signal
        """
        self.run_info_signal.emit(self._wheel.is_running())

    def _status_check(self) -> None:
        """
        Gather the status report of the underlying wheel and send it as a signal.

        Notes
        -----
        Strongly coupled to the exact status report construction of the SteeringWheel
        """
        report = self._wheel.get_status_report()
        if report != self._status_report:
            self._status_report = report
            for k, v in self._status_report.items():
                index = int(k.split(":")[0].strip())
                self.status_signal.emit(index, v)
            if not self._wheel.is_running():
                self.save()

    def _selection_check(self) -> None:
        """
        Queries results of the underlying wheel and send the current selection as a signal.

        Raises
        ------
        RuntimeError
            If the result retrieval fails
        """
        results = self._wheel.get_results()
        if results is None:
            return
        last_selection_result_index = max([i for i, r in enumerate(results)
                                           if r is not None and isinstance(r, SelectionResult)], default=None)
        if last_selection_result_index is not None and last_selection_result_index > self._last_selection_index:
            self._last_selection_index = last_selection_result_index
            result = results[self._last_selection_index]
            if result is None or not isinstance(result, SelectionResult) \
                    or len(self._wheel) <= last_selection_result_index:
                raise RuntimeError("No results available or wrong type")
            name = f"{last_selection_result_index}: {self._wheel.scheme[last_selection_result_index].name}"
            self.selection_signal.emit(name, result)

    def update_step_options(self, settings: dict, index: int) -> None:
        """
        Updates the options of one step in der underlying wheel.

        Parameters
        ----------
        settings : dict
            The new options
        index : int
            The index of the step
        """
        if self._wheel.is_running():
            self.error_signal.emit("Exploration running, ignoring given settings")
            return
        try:
            self._wheel.scheme[index].options.__dict__.update(settings)
        except BaseException as e:
            self.error_signal.emit(f"Failed to transfer settings: {str(e)}")
        self._wheel.save()

    def append(self, exploration_step: ExplorationSchemeStep) -> None:
        """
        Add an exploration step to the wheel.
        """
        self._wheel += exploration_step

    def pop(self, index: int) -> None:
        """
        Remove a step based on index.
        Negative indexing is supported, slicing is not supported.
        """
        self._wheel.pop(index)

    def set_global_selection(self, selection: Selection, for_first: bool) -> None:
        """
        Set a global selection to the wheel.

        Parameters
        ----------
        selection : Selection
            The global selection
        for_first : bool
            If the global selection should be added to the first step.
        """
        self._wheel.set_global_selection(selection, for_first)

    def remove_global_selection(self, selection: Selection) -> None:
        """
        Remove an already set global selection.

        Parameters
        ----------
        selection : Selection
            The already set global selection
        """
        self._wheel.remove_global_selection(selection)

    def get_scheme(self) -> List[ExplorationSchemeStep]:
        """
        Getter for the underlying exploration protocol.

        Returns
        -------
        List[ExplorationResult]
            The exploration protocol
        """
        return self._wheel.scheme

    def get_results(self) -> List[ExplorationResult]:
        """
        Retrieve results and status report from the wheel

        Returns
        -------
        List[ExplorationResult]
            The exploration protocol
        """
        self._wheel.get_status_report()  # to ensure both members are up-to-date
        return self._wheel.get_results()

    def get_partial_restart_info(self) -> Optional[RestartPartialExpansionInfo]:
        return self._wheel.get_partial_restart_info()

    def set_partial_restart_info(self, restart_info: Optional[RestartPartialExpansionInfo]) -> None:
        self._wheel.set_partial_restart_info(restart_info)

    def __len__(self) -> int:
        """
        The length of the wheel (the length of its exploration protocol)
        """
        return len(self._wheel)

    def __bool__(self) -> bool:
        """
        Whether the wheel has an exploration protocol
        """
        return bool(self._wheel)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

from collections import UserList
from typing import Optional, Dict, List, Iterable, Tuple

from PySide2.QtWidgets import (
    QWidget,
    QLabel,
    QLineEdit,
    QTreeView,
)
from PySide2.QtCore import QObject, Qt
from PySide2.QtGui import QStandardItem, QStandardItemModel, QKeySequence

from scine_utilities import opt_settings_names

from scine_heron.containers.buttons import TextPushButton
from scine_heron.containers.combo_box import BaseBox
from scine_heron.containers.layouts import HorizontalLayout, VerticalLayout
from scine_heron.readuct.task_widget import TaskWidget
from scine_heron.utilities import write_error_message, write_info_message
import scine_heron.config as config
from scine_heron.styling.delegates import CustomLightDelegate


class TreeView(QTreeView):
    def __init__(self) -> None:
        super().__init__()
        if config.MODE in config.LIGHT_MODES:
            self.setItemDelegate(CustomLightDelegate())


class ReaductTaskComboBox(BaseBox):
    def __init__(self, parent: Optional[QObject], task_names: List[str]) -> None:
        super().__init__(parent)

        self.setView(TreeView())
        self.setModel(QStandardItemModel())
        self.view().setHeaderHidden(True)
        self.view().setItemsExpandable(False)
        self.view().setRootIsDecorated(False)

        sorted_names = sorted(task_names)
        for name in sorted_names:
            item = QStandardItem(name)
            item.setEnabled(True)
            self.model().appendRow(item)

        self.setCurrentText(sorted_names[0])


class CreateReaductTaskWidget(QWidget):
    default_name: str = "default_name"

    # keys: box name,
    # values:
    #   readuct callable task name if surrounded with "run_{}_task"
    #   number of inputs
    #   number of outputs
    task_mapping: Dict[str, Tuple[str, int, int]] = {
        "Single Point Calculation": ("sp", 1, 1),
        "Structure Optimization": ("opt", 1, 1),
        "Hessian": ("hessian", 1, 1),
        "Transition State Optimization": ("tsopt", 1, 1),
        "Newton Trajectory": ("nt", 1, 1),
        "Newton Trajectory 2": ("nt2", 1, 1),
        "Intrinsic Reaction Coordinate (IRC)": ("irc", 1, 2),
        "B-Spline Transition State Optimization": ("bspline", 2, 1),
        "Artificial Force Induced Reaction (AFIR)": ("afir", 1, 1),
        "AFIR Optimization": ("afir", 1, 1),
        "BSpline Interpolation": ("bspline", 2, 1),
        "Bond Order Calculation": ("bond_order", 1, 1),
        "Geometry Optimization": ("opt", 1, 1),
        "Hessian Calculation": ("hessian", 1, 1),
        "IRC Optimizations": ("irc", 1, 2),
        "NT2 Optimization": ("nt2", 1, 1),
        "NT1 Optimization": ("nt", 1, 1),
        "TS Optimization": ("tsopt", 1, 1)
    }

    def __init__(self, parent: QWidget) -> None:
        super().__init__(parent)
        self._parent = parent
        self.name_edit = QLineEdit(self.default_name)
        name_layout = HorizontalLayout([QLabel("Name:"), self.name_edit])

        self.inputs_edit = MultiHComboBox(self)
        self._input_widget = self.inputs_edit.widget
        inputs_layout = HorizontalLayout([QLabel("Inputs:"), self._input_widget])

        self.outputs_edit = MultiHLineEdit(self)
        outputs_layout = HorizontalLayout([QLabel("Outputs:"), self.outputs_edit])

        self.task_box = ReaductTaskComboBox(self, list(self.task_mapping.keys()))
        self.task_box.setCurrentText("Structure Optimization")
        box_sub_layout = HorizontalLayout([QLabel("Select type:"), self.task_box])

        self.button_add = TextPushButton("Add task", self._generate_new_widget)
        self.button_add.setShortcut(QKeySequence("Return"))

        self.task_box.currentTextChanged.connect(  # pylint: disable=no-member
            self._determine_io_possibilities
        )
        self._determine_io_possibilities()  # call ourselves here once to evaluate default

        self._layout = VerticalLayout()
        self._layout.setAlignment(Qt.AlignTop)
        self._layout.add_layouts([
            box_sub_layout,
            name_layout,
            inputs_layout,
            outputs_layout,
        ])
        self._layout.addWidget(self.button_add)
        self.setLayout(self._layout)

    def add_possible_input_system(self, name: str) -> None:
        self.inputs_edit.append(name)

    def _determine_name(self, cls_name: str) -> str:
        name = self.name_edit.text()
        if name == self.default_name:
            return cls_name
        return name

    def _determine_io_possibilities(self) -> None:
        n_in, n_out = self.task_mapping[self.task_box.currentText()][1:]
        if n_in < 2:
            self.inputs_edit.hide_second()
        else:
            self.inputs_edit.show_second()
        if n_out < 2:
            self.outputs_edit.hide_second()
        else:
            self.outputs_edit.show_second()

    def add_new_possible_system(self, name: str) -> None:
        if name not in self.inputs_edit:
            self.inputs_edit.append(name)

    def _get_settings_suggestions(self) -> Optional[List[str]]:
        if self.task_box.currentText() == "Artificial Force Induced Reaction (AFIR)":
            return self._construct_settings_names(opt_settings_names.Afir) + \
                self._construct_settings_names(opt_settings_names.Convergence)
        if self.task_box.currentText() == "B-Spline Transition State Optimization":
            return None
        if self.task_box.currentText() == "Hessian":
            return ["temperature", "pressure"]
        if self.task_box.currentText() == "Intrinsic Reaction Coordinate (IRC)":
            return self._construct_settings_names(opt_settings_names.Irc) + \
                self._construct_settings_names(opt_settings_names.Convergence)
        if self.task_box.currentText() == "Newton Trajectory":
            return self._construct_settings_names(opt_settings_names.Nt) + \
                self._construct_settings_names(opt_settings_names.Convergence)
        if self.task_box.currentText() == "Newton Trajectory 2":
            return self._construct_settings_names(opt_settings_names.Nt2) + \
                self._construct_settings_names(opt_settings_names.Convergence)
        if self.task_box.currentText() == "Single Point Calculation":
            return ["require_charges", "require_gradients", "require_stress_tensor", "require_bond_orders",
                    "orbital_energies", "silent_stdout_calculator", "spin_propensity_check"]
        if self.task_box.currentText() == "Structure Optimization":
            return self._construct_settings_names(opt_settings_names.Convergence) + \
                self._construct_settings_names(opt_settings_names.Bfgs) + \
                self._construct_settings_names(opt_settings_names.Lbfgs) + \
                self._construct_settings_names(opt_settings_names.NewtonRaphson) + \
                self._construct_settings_names(opt_settings_names.SteepestDescent) + \
                self._construct_settings_names(opt_settings_names.GeometryOptimizer) + \
                self._construct_settings_names(opt_settings_names.CellOptimizer)
        if self.task_box.currentText() == "Transition State Optimization":
            return self._construct_settings_names(opt_settings_names.Convergence) + \
                self._construct_settings_names(opt_settings_names.Bofill) + \
                self._construct_settings_names(opt_settings_names.EigenvectorFollowing) + \
                self._construct_settings_names(opt_settings_names.Dimer) + \
                self._construct_settings_names(opt_settings_names.GeometryOptimizer)
        return None

    @staticmethod
    def _construct_settings_names(obj: object) -> List[str]:
        return [getattr(obj, k) for k in vars(obj).keys() if not k.startswith("_")]

    def _generate_new_widget(self) -> None:
        inputs = self.inputs_edit.get_values()
        if not inputs:
            write_error_message("Must give input systems")
            return
        outputs = self.outputs_edit.get_values()
        task_settings = {}
        if any(out for out in outputs):
            task_settings['output'] = outputs
            self.inputs_edit.extend(outputs)
        else:
            task_settings['output'] = inputs
            write_info_message("No outputs given, using inputs as outputs")

        new_widget = TaskWidget(
            parent=self._parent,
            inputs=inputs,
            task_name=self.task_mapping[self.task_box.currentText()][0],
            widget_title=self._determine_name(self.task_box.currentText()),
            task_settings=task_settings,
            settings_suggestions=self._get_settings_suggestions(),
        )
        self._parent.add_widget(new_widget)


class MultiHComboBox(UserList):

    def __init__(self, parent: Optional[QObject]):
        super().__init__()
        self.widget = QWidget(parent)
        self._first = BaseBox()
        self._second = BaseBox()
        self._layout = HorizontalLayout([self._first, self._second])
        self.widget.setLayout(self._layout)

    def hide_second(self) -> None:
        self._second.setVisible(False)

    def show_second(self) -> None:
        self._second.setVisible(True)

    def get_values(self) -> List[str]:
        if self._second.isVisible():
            return [self._first.currentText(), self._second.currentText()]
        return [self._first.currentText()]

    def _update_impl(self):
        if hasattr(self, "_first") and hasattr(self, "_second"):
            for w in [self._first, self._second]:
                w.clear()
                w.addItems(self.data)

    def __setattr__(self, key, value):
        super().__setattr__(key, value)
        if "data" in key:
            self._update_impl()

    def __delitem__(self, key):
        super().__delitem__(key)
        self._update_impl()

    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        self._update_impl()

    def append(self, item: str) -> None:
        if not item or item in self.data:
            return
        super().append(item)
        self._update_impl()

    def extend(self, other: Iterable[str]) -> None:
        for item in other:
            if not item or item not in self.data:
                super().append(item)
        self._update_impl()

    def insert(self, i: int, item: str) -> None:
        super().insert(i, item)
        self._update_impl()

    def pop(self, i: int) -> str:  # type: ignore  # pylint: disable=signature-differs
        res = super().pop(i)
        self._update_impl()
        return res

    def clear(self) -> None:
        super().clear()
        self._update_impl()

    def remove(self, item: str) -> None:
        super().remove(item)
        self._update_impl()


class MultiHLineEdit(QWidget):

    def __init__(self, parent: Optional[QObject]):
        super().__init__(parent)
        self._first = QLineEdit()
        self._second = QLineEdit()
        self._layout = HorizontalLayout([self._first, self._second])
        self.setLayout(self._layout)

    def hide_second(self) -> None:
        self._second.setVisible(False)

    def show_second(self) -> None:
        self._second.setVisible(True)

    def get_values(self) -> List[str]:
        if self._second.isVisible():
            return [self._first.text().strip(), self._second.text().strip()]
        return [self._first.text().strip()]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
from typing import List, Any

from vtk import (
    vtkMolecule,
    vtkArrayData,
    vtkMoleculeMapper,
    vtkTrivialProducer,
    vtkArray,
    VTK_INT,
    vtkActor,
    vtkInformationVector,
)

from vtk.util.vtkAlgorithm import VTKPythonAlgorithmBase, vtkAlgorithm


class NonEmptyMoleculeError(Exception):
    pass


class AtomSelectionAlgorithm(VTKPythonAlgorithmBase):  # type: ignore[misc]
    """
    Creates a "ghost" molecule
    out of a selection of atoms in the original molecule
    """

    def __init__(self) -> None:
        super().__init__(
            nInputPorts=2,
            nOutputPorts=1,
            inputType=["vtkMolecule", "vtkArrayData"],
            outputType="vtkMolecule",
        )
        self.selected_atom_ids: List[int] = []

    def FillInputPortInformation(self, port: int, info: Any) -> int:
        """Sets the required input type to InputType."""
        info.Set(vtkAlgorithm.INPUT_REQUIRED_DATA_TYPE(), self.InputType[port])
        return 1

    def RequestData(
        self,
        _: Any,
        in_info: List[vtkInformationVector],
        out_info: vtkInformationVector,
    ) -> int:
        """
        Performs the transformation
        """

        base_molecule = vtkMolecule.GetData(in_info[0])
        selection_indices_data = vtkArrayData.GetData(in_info[1])
        selection_indices = selection_indices_data.GetArray(0)
        selection = [
            selection_indices.GetValue(i)
            for i in range(
                selection_indices.GetExtent(0).GetBegin(),
                selection_indices.GetExtent(0).GetEnd(),
            )
        ]

        out_molecule = vtkMolecule.GetData(out_info)

        self.filter_molecule(base_molecule, selection, out_molecule)

        # return True (as an integer) to signal success
        return 1

    @staticmethod
    def filter_molecule(
        base_molecule: vtkMolecule, selection: List[int], out_molecule: vtkMolecule
    ) -> None:
        """
        Modifies the out_molecule in place,
        """
        if out_molecule.GetNumberOfAtoms() > 0:
            raise NonEmptyMoleculeError()

        max_atomic_number = max(
            (
                base_molecule.GetAtom(i).GetAtomicNumber()
                for i in range(base_molecule.GetNumberOfAtoms())
            ),
            default=1,
        )

        for atom_index in selection:
            atom = base_molecule.GetAtom(atom_index)
            new_atom = out_molecule.AppendAtom()

            # The radii of displayed atoms may be modified when values
            # are displayed via the radii. To make sure that selection
            # is visible we use the maximum atomic number for selection.
            new_atom.SetAtomicNumber(max_atomic_number)
            new_atom.SetPosition(atom.GetPosition())


class AtomSelection:
    def __init__(self) -> None:
        """
        The `atom_selector` algorithm has two inputs,
        given by two producers:
        - the molecule
        - the selection

        Pipeline Scheme:

        molecule -> producer ------------------------------> atom_selector -> mapper -> actor
        selection -> selection_data -> selection_producer/
        """

        # Molecule
        self.__producer = vtkTrivialProducer()
        self.__selection_producer = vtkTrivialProducer()

        # Atom selector
        self.__atom_selector = AtomSelectionAlgorithm()
        self.__selection = vtkArray.CreateArray(vtkArray.DENSE, VTK_INT)
        self.__selection.Resize(1)
        self.__selection_data = vtkArrayData()

        # Mapper
        self.__mapper = vtkMoleculeMapper()
        self.__mapper.CreateDefaultLookupTable()
        self.__mapper.SetAtomColorMode(vtkMoleculeMapper.SingleColor)
        self.__mapper.SetAtomicRadiusScaleFactor(0.35)  # default is 0.3

        # Actor
        self.__actor = vtkActor()

        # connecting all together
        self.__selection_data.AddArray(self.__selection)
        self.__selection_producer.SetOutput(self.__selection_data)
        self.__atom_selector.SetInputConnection(0, self.__producer.GetOutputPort())
        self.__atom_selector.SetInputConnection(
            1, self.__selection_producer.GetOutputPort()
        )

        self.__mapper.SetInputConnection(self.__atom_selector.GetOutputPort())
        self.__actor.SetMapper(self.__mapper)
        self.__actor.GetProperty().SetOpacity(0.5)
        self.__actor.GetProperty().SetColor(0.5, 0.5, 0.5)

    def back_to_default_color(self):
        self.__actor.GetProperty().SetColor(0.5, 0.5, 0.5)

    def change_color(self, r: float, g: float, b: float):
        self.__actor.GetProperty().SetColor(r, g, b)

    @property
    def actor(self) -> vtkActor:
        """
        Returns the label actor.
        """
        return self.__actor

    def set_molecule(self, molecule: vtkTrivialProducer) -> None:
        self.__producer.SetOutput(molecule)
        self.set_selection([])

    def get_selection_data(self) -> List[int]:
        data = []
        for i in range(self.__selection.GetSize()):
            data.append(self.__selection.GetValue(i))
        return data

    def set_selection(self, selection: List[int]) -> None:
        self.__selection.Resize(len(selection))
        for i, n in enumerate(selection):
            self.__selection.SetValue(i, n)
        self.__selection_data.Modified()

    def __len__(self):
        return self.__selection.GetSize()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""


from warnings import warn
warn('This is deprecated, please import the queries from scine_database', DeprecationWarning, stacklevel=2)


def finished_calculations():
    return {"status": {"$in": ["complete", "failed", "analyzed"]}}


def unstarted_calculations():
    return {"status": {"$in": ["new", "hold"]}}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

from abc import abstractmethod
from typing import Any, Optional, Union

from PySide2.QtWidgets import (
    QWidget,
    QVBoxLayout,
    QHBoxLayout,
    QLabel,
    QCheckBox,
    QLineEdit,
    QPushButton,
    QGridLayout,
)
from PySide2.QtCore import Qt
from scine_database import Manager
from scine_utilities import ValueCollection

from scine_chemoton.engine import Engine
from scine_chemoton.gears.elementary_steps import ElementaryStepGear
from scine_chemoton.steering_wheel.selections.input_selections import ScineGeometryInputSelection

from scine_heron.chemoton.chemoton_widget import (
    EngineWidget,
    SelectionWidget,
    StepWidget,
    generate_gear_name_settings_suggestions
)
from scine_heron.chemoton.chemoton_widget_container import ChemotonWidgetContainer
from scine_heron.chemoton.class_searcher import ChemotonClassSearcher
from scine_heron.chemoton.grouped_combo_box import GroupedComboBox
from scine_heron.containers.buttons import TextPushButton
from scine_heron.io.text_box import yes_or_no_question
from scine_heron.molecular_viewer import get_mol_viewer_tab
from scine_heron.settings.class_options_widget import \
    generate_instance_based_on_potential_widget_input, ClassOptionsWidget
from scine_heron.utilities import write_error_message
from scine_heron.chemoton.filter_builder import FilterBuilder


class CreateChemotonWidget(QWidget):
    """
    An abstract base class for widgets that allow to build a new widget that holds
    a Chemoton class.

    Notes
    -----
    This widget has hard-coded sizes
    """
    label: str = "Add class"
    default_name: str = "default_name"
    chemoton_class_label = QLabel("Gear:")
    chemoton_class: GroupedComboBox

    def __init__(
            self, parent: ChemotonWidgetContainer, db_manager: Manager, class_searcher: ChemotonClassSearcher
    ) -> None:
        """
        Construct the widget with database information and class that holds the possible classes.

        Parameters
        ----------
        parent : ChemotonWidgetContainer
            The parent widget that holds ChemotonWidgets.
        db_manager : Manager
            The Scine DB Manager that holds the database information
        class_searcher : ChemotonClassSearcher
            A class that holds all underlying classes of the widget this widget
            should be able to construct.
        """
        super().__init__(parent)
        self.db_manager = db_manager
        self.class_searcher = class_searcher
        self.button_add = TextPushButton(self.label, self._generate_new_widget, self)

        self._layout = QVBoxLayout()
        self._layout.setAlignment(Qt.AlignTop)
        self._layout.addWidget(QLabel(self.label))
        self._layout.addWidget(self.button_add)

        name_layout = QHBoxLayout()
        name_layout.addWidget(QLabel("Name:"))
        self.name_edit = QLineEdit(self.default_name)
        name_layout.addWidget(self.name_edit)
        self._layout.addLayout(name_layout)

        self._add_chemoton_class_at_layout()

        self.filter_check: Optional[QCheckBox] = None
        self.settings_check: Optional[QCheckBox] = None
        self._layout_additions()

        self.setLayout(self._layout)

        self.setMaximumWidth(500)
        self.setMinimumWidth(300)
        self.setMaximumHeight(400)
        self.setMinimumHeight(200)

    def _add_filter_checkbox_to_layout(self):
        """
        Adds a checkbox to our layout with the question if we want to include the
        filter of the builder.
        """
        layout = QHBoxLayout()
        self.filter_check = QCheckBox("Take Filter from Builder")
        self.filter_check.setChecked(False)
        layout.addWidget(self.filter_check)
        self._layout.addLayout(layout)

    def _add_settings_checkbox_to_layout(self):
        """
        Adds a check box to our layout with the question if we want to include the
        settings of the builder.
        """
        layout = QHBoxLayout()
        self.settings_check = QCheckBox("Take Settings from Builder")
        self.settings_check.setChecked(False)
        layout.addWidget(self.settings_check)
        self._layout.addLayout(layout)

    def _add_chemoton_class_at_layout(self) -> None:
        """
        Adds a combox box that allows to select the underlying Chemoton class
        and the surrounding name information and selections.
        """
        layout = QHBoxLayout()

        self.chemoton_class = GroupedComboBox(self, self.class_searcher)

        layout.addWidget(self.chemoton_class_label)
        layout.addWidget(self.chemoton_class)
        self._layout.addLayout(layout)

    def _filters_are_activated(self) -> bool:
        """
        Returns
        -------
        bool
            If the filters checkbox exists and is ticked.
        """
        return self.filter_check is not None and self.filter_check.isChecked()

    def _determine_name(self, cls_name: str) -> str:
        """
        Determines the name of the class based on the given name. If the name is the
        default name, we return the classes name.

        Parameters
        ----------
        cls_name : str
            The entered class name

        Returns
        -------
        str
            The class name to be given to the new widget.
        """
        name = self.name_edit.text()
        if name == self.default_name:
            return cls_name
        return name

    @abstractmethod
    def _layout_additions(self) -> None:
        """
        Possible additional layout additions that can be implemnted by child class
        """

    @abstractmethod
    def _generate_new_widget(self) -> None:
        """
        The generation of a new widget.
        """


class CreateEngineWidget(CreateChemotonWidget):
    """
    Widget that creates and adds a new widget that holds and controls a Chemoton engine.
    """
    label: str = "Add Engine"
    chemoton_class_label = QLabel("Gear:")

    def _layout_additions(self) -> None:
        """
        Adds the filter and settings check boxes to the layout
        """
        self._add_filter_checkbox_to_layout()
        self._add_settings_checkbox_to_layout()

    def _generate_new_widget(self) -> None:
        """
        Construct the engine based on the selected text in the combobox together
        with the built filters and settings, if wanted. Also carries out the complete
        settings mapping for more complex gear option structures.

        Notes
        -----
        Actual construction of the widgets is done by calling the `add_widget`
        method of our parent.
        """
        gear_name = self.chemoton_class.currentText().strip()

        credentials = self.db_manager.get_credentials()
        engine = Engine(credentials)
        gear = self.class_searcher[gear_name]()

        # take care of model
        if hasattr(gear.options, "model"):
            model = self.parent().get_model()
            setattr(gear.options, "model", model)

        # take care of filters
        filter_description = ""
        if self._filters_are_activated():
            filters = self.parent().get_filters()
            filter_names = [
                "aggregate_filter",
                "reactive_site_filter",
                "further_exploration_filter",
                "structure_filter",
                "elementary_step_filter",
                "reaction_filter",
                "aggregate_enabling",
                "aggregate_validation",
                "reaction_enabling",
                "result_enabling",
                "reaction_validation",
                "reaction_disabling",
                "step_disabling"
            ]
            assert len(filters) == len(filter_names)

            for name, filter_inst in zip(filter_names, filters):
                was_set = False
                # All filters are handled in the same way except the reactive site and further exploration filter.
                # These are not a members of the gear itself but of the trial generator of the elementary step gears.
                if name not in ["reactive_site_filter", "further_exploration_filter"]:
                    if hasattr(gear, name):
                        setattr(gear, name, filter_inst)
                        was_set = True
                else:
                    if hasattr(gear, "trial_generator") and hasattr(getattr(gear, "trial_generator"), name):
                        setattr(getattr(gear, "trial_generator"), name, filter_inst)
                        was_set = True
                if was_set:
                    filter_description += f"{filter_inst.name}\n"
        if not filter_description:
            filter_description = "None"

        # take care of settings
        if self.settings_check is not None and self.settings_check.isChecked():
            settings = self.parent().get_settings()

            def evaluate_new_value(old_value: Any) -> Union[ValueCollection, dict, bool]:
                if isinstance(old_value, ValueCollection):
                    return settings if not old_value else ValueCollection({**old_value.as_dict(), **settings.as_dict()})
                elif isinstance(old_value, dict):
                    return settings.as_dict() if not old_value else {**old_value, **settings.as_dict()}
                return False

            attr, mapping = EngineWidget.get_settings(gear)
            for k, v in attr.items():
                if "settings" in k:
                    new_value = evaluate_new_value(v)
                    if not new_value:
                        continue
                    if hasattr(gear.options, k):
                        setattr(gear.options, k, new_value)
                    if isinstance(gear, ElementaryStepGear):
                        if hasattr(gear.trial_generator.options, k):
                            setattr(gear.trial_generator.options, k, new_value)
                        if "." in k:
                            mapping_key = k.split(".")[0]
                            if mapping_key in mapping:
                                for subkey, subvalue in mapping[mapping_key].items():
                                    if "settings" in subkey:
                                        unmod_key = subkey.split(".")[-1]
                                        new_value = evaluate_new_value(subvalue)
                                        if not new_value:
                                            continue
                                        setattr(getattr(gear.trial_generator.options, mapping_key),
                                                unmod_key, new_value)

        engine.set_gear(gear)
        new_widget = EngineWidget(
            self.parent(),
            engine,
            gear,
            self._determine_name(gear_name),
            gear_name,
            filter_description
        )
        self.parent().add_widget(new_widget)


class CreateStep(CreateChemotonWidget):
    """
    Widget that creates and adds a new widget that holds and controls a Network Expansion.
    """
    label: str = "Add Network Expansion"
    chemoton_class_label = QLabel("Network Expansion:")

    def _layout_additions(self) -> None:
        """
        Adds the settings checkbox to the layout
        """
        self._add_settings_checkbox_to_layout()

    def _generate_new_widget(self) -> None:
        """
        Generates the underlying Network Expansion with the set model, optionally
        the built settings and some possible user input via a pop-up.

        Notes
        -----
        Actual construction of the widgets is done by calling the `add_widget`
        method of our parent.
        """
        step_name = self.chemoton_class.currentText().strip()

        model = self.parent().get_model()
        predefined_kwargs = {
            "model": model,
        }
        if self.settings_check is not None and self.settings_check.isChecked():
            settings = self.parent().get_settings()
            predefined_kwargs["general_settings"] = settings

        step_cls = self.class_searcher[step_name]
        suggestions = generate_gear_name_settings_suggestions()
        step = generate_instance_based_on_potential_widget_input(self, step_cls, predefined_kwargs, suggestions)

        new_widget = StepWidget(
            self.parent(),
            step,
            self._determine_name(step_name),
            step_name,
        )
        self.parent().add_widget(new_widget)


class CreateSelection(CreateChemotonWidget):
    """
    Widget that creates and adds a new widget that holds and controls a Selection.
    """
    label: str = "Add Selection Step"
    chemoton_class_label = QLabel("Selection Step:")

    def _layout_additions(self) -> None:
        """
        Adds the filter checkbox and everything related to the global selection logic
        to the layout
        """
        self._add_filter_checkbox_to_layout()
        # global actual widgets
        self._button_add_global = QPushButton("Add as global Selection")
        self._button_add_global.clicked.connect(  # pylint: disable=no-member
            lambda: self._generate_new_widget(as_global=True)
        )
        self._global_first_check = QCheckBox("Apply global selection\nfor first step")
        self._global_first_check.setChecked(False)
        # global widgets layout design
        self._global_container_widget = QWidget()
        global_grid = QGridLayout()
        global_grid.addWidget(self._button_add_global, 0, 0)
        global_grid.addWidget(self._global_first_check, 0, 1)
        self._global_container_widget.setLayout(global_grid)
        self._layout.addWidget(self._global_container_widget)

    def _generate_new_widget(self, as_global: bool = False) -> None:
        """
        Constructs the underlying selection with the set model and built filters.
        Also handles special case of ScineGeometryInputSelection and writes the filter
        description.

        Parameters
        ----------
        as_global : bool, optional
            If the built selection is global, by default False

        Notes
        -----
        Actual construction of the widgets is done by calling the `add_widget`
        method of our parent.
        """

        selection_name = self.chemoton_class.currentText().strip()

        model = self.parent().get_model()
        predefined_kwargs = {
            "model": model
        }
        if self._filters_are_activated():
            filters = self.parent().get_filters()
            predefined_kwargs["additional_aggregate_filters"] = [filters[0]]
            predefined_kwargs["additional_reactive_site_filters"] = [filters[1]]
        else:
            predefined_kwargs["additional_aggregate_filters"] = None
            predefined_kwargs["additional_reactive_site_filters"] = None
        selection_cls = self.class_searcher[selection_name]

        # special case for mol viewer selection
        if selection_cls == ScineGeometryInputSelection:
            """
            this code block requires some updates in other modules
            answer = yes_or_no_question(self, "Do you want to take the current calculator from the molecular viewer")
            if answer:
                tab = get_mol_viewer_tab(want_atoms_there=True)
                if tab is None or tab.mol_widget is None:
                    # error was printed
                    return
                calculator = tab.create_calculator_widget.get_calculator()
                predefined_kwargs["structure_information_input"] = [calculator]
            else:
            """
            answer2 = yes_or_no_question(self,
                                         "Do you want to take the current structure from the molecular viewer")
            if answer2:
                tab = get_mol_viewer_tab(want_atoms_there=True)
                if tab is None or tab.mol_widget is None:
                    # error was printed
                    return
                atoms = tab.mol_widget.get_atom_collection()
                required_input = {
                    "molecular_charge": 0,
                    "spin_multiplicity": 1,
                }
                inp = ClassOptionsWidget(required_input, parent=self, allow_additions=False, allow_removal=False)
                inp.exec_()
                predefined_kwargs["structure_information_input"] = [
                    (atoms, required_input["molecular_charge"], required_input["spin_multiplicity"])
                ]
            else:
                write_error_message(f"Abort {ScineGeometryInputSelection.__name__} creation")
                return

        selection = generate_instance_based_on_potential_widget_input(self, selection_cls, predefined_kwargs)

        default_filters = FilterBuilder.default_filters()
        filter_description = ""
        if self._filters_are_activated():
            for default, filter_inst in zip(default_filters, filters):
                if filter_inst is not default:
                    filter_description += f"{filter_inst.name}\n"
        if not filter_description:
            filter_description = "None"

        new_widget = SelectionWidget(
            self.parent(),
            selection,
            self._determine_name(selection_name),
            selection_name,
            filter_description,
            as_global=as_global,
            global_for_first=self._global_first_check.isChecked()
        )
        self.parent().add_widget(new_widget, predefined_position=as_global)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

from scine_art.database import ReactionTemplateDatabase
from scine_art.reaction_template import ReactionTemplate

from typing import TYPE_CHECKING, Any, List, Optional
from PySide2.QtCore import QObject
if TYPE_CHECKING:
    Signal = Any
else:
    from PySide2.QtCore import Signal


class ReactionTemplateStorage(QObject):

    reaction_template_count_changed = Signal(int)

    def __init__(self) -> None:
        super().__init__()
        self.__template_db = ReactionTemplateDatabase()
        self.__template_ids: List[str] = []

    def add_reaction_template(self, new_template: ReactionTemplate) -> None:
        new_id = self.__template_db.add_template(new_template)
        if new_id not in self.__template_ids:
            self.__template_ids.append(new_id)
            self.reaction_template_count_changed.emit(self.__template_db.template_count())
        assert self.__template_db.template_count() == len(self.__template_ids)

    def clear(self):
        self.__template_db = ReactionTemplateDatabase()
        self.__template_ids = []
        assert self.__template_db.template_count() == len(self.__template_ids)
        self.reaction_template_count_changed.emit(self.__template_db.template_count())

    def load_database(self, filename: str) -> None:
        self.clear()
        self.add_database(filename)

    def add_database(self, filename: str) -> None:
        self.__template_db.append_file(filename)
        self.__template_ids = []
        for rt in self.__template_db.iterate_templates():
            self.__template_ids.append(rt.get_uuid())
        self.reaction_template_count_changed.emit(self.__template_db.template_count())

    def save_database(self, filename: str) -> None:
        self.__template_db.save(filename)

    def get_template_ids(self) -> List[str]:
        return self.__template_ids

    def get_template(self, rt_id: str) -> Optional[ReactionTemplate]:
        return self.__template_db.get_template(rt_id)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
"""
Provides the MoleculeLabels class.
"""

from typing import Any, List, Optional

from vtk import (
    vtkActor2D,
    vtkInformationVector,
    vtkLabeledDataMapper,
    vtkMolecule,
    vtkPeriodicTable,
    vtkPoints,
    vtkPolyData,
    vtkStringArray,
    vtkTrivialProducer,
    vtkSelectVisiblePoints,
    vtkAlgorithm,
)
from vtk.util.vtkAlgorithm import VTKPythonAlgorithmBase

from scine_heron.settings.settings import LabelsStyle


class LabelGenerator:
    """
    Converts atoms to labels based on a given style.
    """

    def __init__(self, style: LabelsStyle = LabelsStyle.Empty):
        self.style = style

    def generate_label(self, molecule: vtkMolecule, atom_index: int) -> str:
        """
        Generates the label for the atom with the given atomic index.
        """
        if self.style == LabelsStyle.Empty:
            array = molecule.GetVertexData().GetArray("scine_labels")
            return "" if array is None else str(array.GetValue(atom_index))

        atom = molecule.GetAtom(atom_index)
        if self.style == LabelsStyle.Symbol:
            return str(vtkPeriodicTable().GetSymbol(atom.GetAtomicNumber()))
        if self.style == LabelsStyle.AtomicNumber:
            return str(atom.GetAtomicNumber())
        if self.style == LabelsStyle.IndexNumber:
            return str(atom_index)
        raise RuntimeError("Unknown label style!")


class MoleculeToLabel(VTKPythonAlgorithmBase):  # type: ignore[misc]
    """
    Convert the input (a vtkMolecule) to the output (a vtkPolyData)
    by transforming each volume to its center and its label.
    """

    def __init__(self, style: LabelsStyle = LabelsStyle.Empty):
        super().__init__(
            nInputPorts=1,
            nOutputPorts=1,
            inputType="vtkMolecule",
            outputType="vtkPolyData",
        )

        self.__generator = LabelGenerator(style)

    def RequestData(
        self,
        _: Any,
        in_info: List[vtkInformationVector],
        out_info: vtkInformationVector,
    ) -> int:
        """
        Performs the transformation.
        """
        molecule = vtkMolecule.GetData(in_info[0])

        number_of_atoms = molecule.GetNumberOfAtoms()

        points = vtkPoints()
        strings = vtkStringArray()
        points.SetNumberOfPoints(number_of_atoms)
        strings.SetNumberOfValues(number_of_atoms)

        for atom_index in range(number_of_atoms):
            atom = molecule.GetAtom(atom_index)
            points.SetPoint(atom_index, atom.GetPosition())
            strings.SetValue(
                atom_index, self.__generator.generate_label(molecule, atom_index)
            )

        data = vtkPolyData.GetData(out_info)
        data.SetPoints(points)
        data.GetPointData().AddArray(strings)

        # return True (as an integer) to signal success
        return 1

    @property
    def style(self) -> LabelsStyle:
        """
        Returns the style of the labels.
        """
        return self.__generator.style

    @style.setter
    def style(self, style: LabelsStyle) -> None:
        """
        Sets the label style and updates the output if necessary.
        """
        if self.__generator.style == style:
            return

        self.__generator.style = style

        # the output of the algorithm needs to be recomputed
        # because different labels will be generated
        self.Modified()


class MoleculeLabels:
    """
    Displays molecule labels in a 3D view.
    """

    def __init__(
        self,
        selector: Optional[vtkSelectVisiblePoints] = None,
        style: LabelsStyle = LabelsStyle.Empty,
    ):
        self.__producer = vtkTrivialProducer()
        self.__labeler = MoleculeToLabel(style)
        mapper = self.__create_mapper()
        self.__connect_pipeline(self.__producer, self.__labeler, selector, mapper)

        self.__actor = self.__create_actor(mapper)

    @staticmethod
    def __create_mapper() -> vtkLabeledDataMapper:
        """
        Creates a mapper that converts strings at points to 2D labels.
        """
        mapper = vtkLabeledDataMapper()
        mapper.SetLabelModeToLabelFieldData()
        mapper.GetLabelTextProperty().SetFontSize(12)
        mapper.GetLabelTextProperty().SetJustificationToCentered()
        mapper.GetLabelTextProperty().SetVerticalJustificationToCentered()

        return mapper

    @staticmethod
    def __create_actor(mapper: vtkLabeledDataMapper) -> vtkActor2D:
        """
        Create 2D actor that displays the atom labels.
        """
        actor = vtkActor2D()
        actor.SetMapper(mapper)

        return actor

    @staticmethod
    def __connect_pipeline(*args: Optional[vtkAlgorithm]) -> None:
        """
        Connects the provided components in a pipeline.
        Ignores None components.
        """
        previous = None
        for arg in args:
            if previous is None:
                previous = arg
                continue

            if arg is None:
                continue

            arg.SetInputConnection(previous.GetOutputPort())
            previous = arg

    @property
    def actor(self) -> vtkActor2D:
        """
        Returns the label actor.
        """
        return self.__actor

    def display_labels_style(self, style: LabelsStyle) -> None:
        """
        Change labels style.
        """
        self.__labeler.style = style

    def set_molecule(self, molecule: vtkMolecule) -> None:
        self.__producer.SetOutput(molecule)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

from typing import Optional
from PySide2.QtWidgets import QLabel
from PySide2.QtCore import QObject


class WrappedLabel(QLabel):
    """
    A QLabel with the word wrap enabled by default.
    """

    def __init__(self, text: str, parent: Optional[QObject] = None):
        """
        Construct the label with the text and activate word wrap.

        Parameters
        ----------
        text : str
            The text in the label
        parent : Optional[QObject], optional
            The parent of the widget, by default None
        """
        super().__init__(text, parent=parent)
        self.setWordWrap(True)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

from typing import List, Tuple, Optional
from copy import deepcopy
from math import nan

from PySide2.QtWidgets import (
    QGridLayout,
    QLabel,
    QPushButton,
    QWidget,
    QRadioButton,
    QHBoxLayout,
    QDoubleSpinBox,
    QVBoxLayout
)

from scine_heron.autocas.autocas_settings import AutocasSettings
from scine_heron.autocas.signal_handler import SignalHandler


class MODiagram(QWidget):
    def __init__(self, parent: QWidget, signal_handler: SignalHandler, settings: AutocasSettings):
        # TODO: make mo diagram scrollable
        QWidget.__init__(self, parent)
        self.__layout = QVBoxLayout()

        self.signal_handler = signal_handler
        self.autocas_settings = settings

        self.buttons: List[Tuple[QRadioButton, QPushButton]] = []

        self.apply_button = QPushButton("Set CAS")
        self.__layout.addWidget(self.apply_button)
        # pylint: disable-next=E1101
        self.apply_button.clicked.connect(self.get_checked)

        self.setLayout(self.__layout)

        # self.signal_handler.update_mo_diagram.connect(self.make_new)
        self.signal_handler.update_mo_diagram.connect(self.update)
        self.signal_handler.update_mo_diagram.connect(self.contour_setter)
        self.signal_handler.update_mo_diagram.connect(self.check_buttons)

        self.__homo_lumo_label: Optional[QWidget] = None
        self.__contour_value_layout: Optional[QHBoxLayout] = None

    # def make_new(self):
    #     self.__layout = QVBoxLayout()
    #     self.setLayout(self.__layout)

    def contour_setter(self):
        self.__contour_value_layout = QHBoxLayout()
        label = QLabel("Contour Value")
        self.__contour_value_layout.addWidget(label)
        box = QDoubleSpinBox()
        box.setValue(0.05)
        self.__contour_value_layout.addWidget(label)
        self.__contour_value_layout.addWidget(box)
        self.__layout.addLayout(self.__contour_value_layout)
        self.setLayout(self.__layout)

    def get_value(self):
        for orb_index, (show, _) in enumerate(reversed(self.buttons)):
            if show.isChecked():
                self.signal_handler.view_orbital.emit(orb_index + 1)
                break

    def check_buttons(self):
        # occupation = self.autocas_settings.user_occupation
        indices = self.autocas_settings.user_indices
        # print(indices)
        # print(occupation)
        if indices is None:
            # print("Note: No autoCAS selection available/read from disk!")
            return
        if self.buttons is None:
            return
        for i, (_, b) in enumerate(self.buttons[::-1]):
            # print(i, occupation[i])
            if i in indices:
                b.setChecked(True)

    def clean_up(self):
        """
        Removes all buttons, the contour value layout and the HOMO-LUMO gap label.
        """
        for s, b in self.buttons:
            s.deleteLater()
            b.deleteLater()
        self.buttons = []
        if self.__homo_lumo_label:
            self.__homo_lumo_label.deleteLater()
        if self.__contour_value_layout:
            for i in reversed(range(self.__contour_value_layout.count())):
                self.__contour_value_layout.itemAt(i).widget().deleteLater()
            self.__contour_value_layout.deleteLater()

    @staticmethod
    def degenerate(e, last_e, g, last_g) -> bool:
        if g != last_g:
            return False
        return abs(e - last_e) < 1e-4

    def update(self):
        # label = QLabel("Molecular Orbitals")
        # self.__layout.addWidget(label)
        layout = QGridLayout()
        self.clean_up()
        button_grid = [[]]
        show_button_grid = [[]]
        last_e = float("inf")
        last_g = nan
        row = 0
        energies = deepcopy(self.autocas_settings.orbital_energies)
        occupations = self.autocas_settings.orbital_occupations
        orbital_index_sets = self.autocas_settings.orbital_index_sets
        group_indices: List[int] = [len(energies) for _ in energies]
        if orbital_index_sets is not None:
            for i_group, indices in enumerate(orbital_index_sets):
                for i_orb in indices:
                    group_indices[i_orb - 1] = i_group
        for i, (e, g) in enumerate(zip(reversed(energies), reversed(group_indices))):
            x = len(energies) - i
            string = "{:n}".format(x)
            if g < len(energies):
                string += "|G" + str(g)

            button_show = QRadioButton()
            button_show.setChecked(False)
            button_show.setMinimumWidth(20)
            button_show.setMaximumWidth(20)
            button_show.setMinimumHeight(20)
            button_show.setMaximumHeight(20)
            # pylint: disable-next=E1101
            button_show.toggled.connect(self.get_value)

            button = QPushButton(string)
            button.setCheckable(True)
            button.setMinimumWidth(90)
            # button.setMaximumWidth(60)
            button.setMinimumHeight(20)
            button.setMaximumHeight(20)
            self.buttons.append((button_show, button))
            if not self.degenerate(e, last_e, g, last_g):
                row += 1
                button_grid.append([button])
                show_button_grid.append([button_show])
            else:
                button_grid[row].append(button)
                show_button_grid[row].append(button_show)
            last_e = e
            last_g = g
        max_cols = 3
        r = 1
        counter = 0
        printed_gap_info = False
        for r_ind, row in enumerate(button_grid):
            modulo_modifier = 0
            for c, col in enumerate(row):
                if occupations[len(energies) - counter - 1] and not printed_gap_info:
                    printed_gap_info = True
                    r += 1
                    # self.layout.addWidget(QHLine(), r, 0, 1, max_cols)
                    r += 1
                    self.__homo_lumo_label = QLabel("HOMO-LUMO-Gap")
                    self.__homo_lumo_label.setMinimumHeight(20)
                    self.__homo_lumo_label.setMaximumHeight(20)
                    layout.addWidget(self.__homo_lumo_label, r, 0, 1, max_cols)
                    r += 1
                    # self.layout.addWidget(QHLine(), r, 0, 1, max_cols)
                    r += 1
                    modulo_modifier = c
                mod = (c - modulo_modifier) % max_cols
                if mod == 0:
                    r += 1
                layout_2 = QHBoxLayout()
                layout_2.setSpacing(10)
                layout_2.setMargin(10)
                layout_2.setContentsMargins(0, 0, 0, 0)
                layout_2.addWidget(show_button_grid[r_ind][c])
                layout_2.addWidget(col)
                layout.addLayout(layout_2, r, mod, 1, 1)
                counter += 1
            r += 1
        self.__layout.addLayout(layout)
        self.setLayout(self.__layout)
        # self.layout.addWidget(QHLine(), r, 0, 1, max_cols)

    def get_checked(self):
        cas_indices = []
        cas_occupation = []
        for i, b in enumerate(self.buttons[::-1]):
            if b.isChecked():
                cas_indices.append(i)
                cas_occupation.append(self.autocas_settings.orbital_occupations[i])
        # print(cas_occupation)
        # print(cas_indices)
        self.autocas_settings.user_occupation = cas_occupation
        self.autocas_settings.user_indices = cas_indices



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

import os
from enum import Enum
from typing import List, Set, Optional


class Stage(Enum):
    InitialOrbitals = 1
    InitialDMRG = 2
    FinalDMRG = 3


class AutocasSettings:
    def __init__(self):
        # not for autocas
        self.molecule_xyz_file = None
        self.molcas_orbital_file = ""
        self.orbital_energies = None
        self.orbital_occupations = None
        # for autocas

        self.molecule_charge = 0
        self.molecule_spin_multiplicity = 1
        self.molecule_double_d_shell = True

        self.plateau_values = 10
        self.threshold_step = 0.01

        self.weak_correlation_threshold = 0.02
        self.single_reference_threshold = 0.14

        self.enable_large_cas = False
        self.large_spaces_max_orbitals = 20

        # self.molecule_core_orbitals = 0
        # self.molecule_valence_orbitals = 0
        # self.molecule_electrons = 0
        # self.molecule_occupation = []

        self.molcas_dump = True
        self.molcas_project_name = "autocas"
        # setup interface

        self.molcas_work_dir = os.getcwd() + "/test"
        # Interface
        self.interface_dmrg_bond_dimension = 250
        self.interface_dmrg_sweeps = 5
        self.interface_basis_set = "cc-pvdz"
        self.interface_method = "dmrg_ci"
        self.interface_post_cas_method = ""
        # self.interface_spin_multiplicity = 1
        # self.interface_charge = 0
        self.interface_xyz_file = None
        self.molcas_point_group = "C1"
        self.molcas_ipea = 0.0
        self.interface_cholesky = False
        self.interface_uhf = False
        self.interface_fiedler = False
        # self.interface_only_hf = True
        self.interface_n_excited_states = 0

        self.result_initial_occupation = None
        self.result_initial_indices = None

        self.result_initial_s1 = None
        self.result_initial_mutual_information = None
        self.result_initial_energy = None

        self.result_final_occupation = None
        self.result_final_indices = None

        self.user_occupation = None
        self.user_indices = None

        self.orbital_index_sets: Optional[List[Set[int]]] = None

    def get_orbital_file(self, stage: Stage = Stage.InitialOrbitals):
        molden_string = self.molcas_work_dir + "/" + self.molcas_project_name + \
            "/initial/" + self.molcas_project_name + ".scf.molden"
        if stage == Stage.InitialOrbitals:
            self.molcas_orbital_file = molden_string
            return molden_string
        # TODO: make this useful
        return molden_string

        # MOLCAS
        # self.molcas_orbital_localisation = False
        # self.molcas_localisation_space = "OCCUpied"
        # self.molcas_localisation_method = "PIPEk-Mezey"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
from time import sleep
from typing import Optional, List, Callable, Any, Union
from threading import Thread

from PySide2.QtCore import QObject
from PySide2.QtGui import QCloseEvent
from PySide2.QtWidgets import (
    QLineEdit,
    QLabel,
    QHBoxLayout,
    QWidget,
    QStyle,
)

from scine_heron.steering.wheel_handling import WheelThread
from scine_database import Manager, Model
from scine_chemoton.steering_wheel.datastructures import (
    SelectionResult,
    Status,
    RestartPartialExpansionInfo
)
from scine_chemoton.steering_wheel.network_expansions import (
    GiveWholeDatabaseWithModelResult
)
from scine_chemoton.steering_wheel.selections import (
    Selection,
    AllCompoundsSelection,
    SelectionAndArray,
    SelectionOrArray,
    PredeterminedSelection,
)

from scine_heron.chemoton.chemoton_widget_container import ChemotonWidgetContainer
from scine_heron.chemoton.chemoton_widget import StepWidget, SelectionWidget, ExplorationStepWidget

from scine_heron.containers.start_stop_widget import StartStopWidget
from scine_heron.io.text_box import yes_or_no_question
from scine_heron.utilities import (
    write_error_message,
    write_info_message,
)


class SteeringDisplay(ChemotonWidgetContainer):
    """
    This class is the main class for the steering display.
    It holds all the subwidgets, handles the communication between them and
    holds the thread that takes care of the Steering Wheel communication.
    It also starts and stops the exploration.
    It inherits from the Chemoton container and burrows most of its functionality.

    Notes
    -----
    So far the SteeringWheel logic and display logic are not well decoupled, hence
    currently still some knowledge about the underlying Chemoton logic is recommended when manipulate things.
    """

    def __init__(self, parent: Optional[QObject], db_manager: Manager, wanted_classes: List[Any],
                 widget_creators: List[Callable], cls_black_list: Optional[List[type]] = None) -> None:
        """
        Construct the tab that holds the steering display.

        Parameters
        ----------
        parent : Optional[QObject]
            The parent of this widget.
        db_manager : Manager
            The database manager, determines the database we are exploring.
        wanted_classes : List[Any]
            The classes we should be able to construct.
        widget_creators : List[Callable]
            The functions that create the given classes.
        cls_black_list : Optional[List[type]], optional
            The classes that should not be creatable, by default None
        """
        steering_black_list = [SelectionAndArray, SelectionOrArray, PredeterminedSelection]
        if cls_black_list is None:
            cls_black_list = steering_black_list
        else:
            cls_black_list += steering_black_list
        # import to only call super here, because we want to pass the black list
        ChemotonWidgetContainer.__init__(self, parent, db_manager, wanted_classes, widget_creators, cls_black_list)

        self._settings_builder.remove_chemoton_button()
        self._max_widgets_per_row = 2
        self._is_currently_deleting = False

        self._wheel_thread = WheelThread(self, self.db_manager.get_credentials())

        self._toolbar.shortened_add_action(QStyle.SP_BrowserReload, "Restart exploration",
                                           "F5", self._restart)
        self._toolbar.shortened_add_action(QStyle.SP_DialogCancelButton, "Terminate exploration",
                                           "Ctrl+K", self._terminate)

        # take care of global selection
        self._global_selection_widget = self._default_global_selection()
        self._global_selection_position = (1, 0)
        self._grid_layout.addWidget(self._global_selection_widget, *self._global_selection_position)

        # take care of run information
        run_info_holder = QWidget()
        run_info_label = QLabel("Exploration status:")
        self._run_info = QLineEdit("not running")
        self._run_info.setReadOnly(True)
        run_info_layout = QHBoxLayout()
        run_info_layout.addWidget(run_info_label)
        run_info_layout.addWidget(self._run_info)
        run_info_holder.setLayout(run_info_layout)
        self._scroll_area_content_layout.addWidget(run_info_holder)

        # take care of Selection Display
        from .current_selection import CurrentSelectionDisplay
        self._current_selection_widget = CurrentSelectionDisplay(self)
        self._current_selection_position = (0, 2)
        self._grid_layout.addWidget(self._current_selection_widget, *self._current_selection_position)

        # take care of signals
        self._wheel_thread.status_signal.connect(  # pylint: disable=no-member
            self.color_widget
        )
        self._wheel_thread.status_signal.connect(  # pylint: disable=no-member
            self.activate_copy
        )
        self._wheel_thread.selection_signal.connect(  # pylint: disable=no-member
            self.construct_new_current_selection_widget
        )
        self._wheel_thread.run_info_signal.connect(  # pylint: disable=no-member
            self._handle_running_status
        )
        self._wheel_thread.error_signal.connect(  # pylint: disable=no-member
            write_error_message
        )
        self._stop_thread: Optional[Thread] = None
        self._wheel_thread.start()

        self.setMinimumWidth(600)
        self.setLayout(self._grid_layout)

    def closeEvent(self, event: QCloseEvent) -> None:
        """
        Takes care of closing existing threads.
        """
        if self._stop_thread is not None:
            self._stop_thread.join()
        self._wheel_thread.terminate_wheel(save_progress=True)
        self._wheel_thread.terminate()
        self._wheel_thread.wait()
        super().closeEvent(event)

    def _handle_running_status(self, is_running: bool) -> None:
        """
        Should be called whenever the running status changes.
        Updates the box and the settings buttons of all subwidgets, because
        we do not want to change the settings while the exploration is running.

        Parameters
        ----------
        is_running : bool
            Our status
        """
        for widget in self.created_chemoton_widgets:
            if widget.button_settings is not None:
                widget.button_settings.setEnabled(not is_running)
            if widget.button_delete is not None:
                widget.button_delete.setEnabled(not is_running)
        if self._global_selection_widget.button_settings is not None:
            self._global_selection_widget.button_settings.setEnabled(not is_running)
        if self._global_selection_widget.button_delete is not None:
            self._global_selection_widget.button_delete.setEnabled(not is_running)
        text = "running" if is_running else "not running"
        self._run_info.setText(text)

    def _pre_io_operations(self) -> None:
        """
        Any operations to be carried out shortly before reading/writing to disk.
        Retrieves the current results from the thread and updates the widgets.

        Raises
        ------
        RuntimeError
            If unknown subwidgets are encountered.
        """
        # makes sure that all results are updated
        results = self._wheel_thread.get_results()
        if results is not None:
            scheme = self._wheel_thread.get_scheme()
            for i, (result, step) in enumerate(zip(results, scheme)):
                widget = self.created_chemoton_widgets[i]
                if not isinstance(widget, StepWidget) and not isinstance(widget, SelectionWidget):
                    raise RuntimeError("Encountered unexpected widget type")
                widget.exploration_step.status = step.status
                widget.exploration_step.set_result(result)
        self._wheel_thread.pause()

    def _post_io_operations(self) -> None:
        """
        Any operations to be carried out shortly after reading/writing to disk.
        """
        self._wheel_thread.resume()

    def _default_global_selection(self) -> SelectionWidget:
        """
        Returns the default global selection widget,
        reconstruct every time to avoid mutability issues.

        Returns
        -------
        SelectionWidget
            The default global selection widget, which is a selection of all compounds.
        """
        widget = SelectionWidget(
            self,
            AllCompoundsSelection(self.get_model()),
            "Global Selection",
            "none",
            "None",
            True,
            True
        )
        widget.setMinimumWidth(300)
        widget.setMaximumWidth(500)
        widget.setMaximumHeight(200)
        widget.setMinimumHeight(200)
        return widget

    def add_widget(self, chemoton_widget: Union[StepWidget, SelectionWidget],
                   predefined_position: bool = False) -> None:
        """
        Overloads the base widget with additional logic based on modifications,
        that the SteeringWheel may do with added steps.

        Possible modifications are:
        - adding a global selection widget
        - adding a current selection widget
        - combining subsequent selection widgets
        - added a whole DB expansion widget before the first selection

        Parameters
        ----------
        chemoton_widget : Union[StepWidget, SelectionWidget]
            The subwidget that is to be added.
        predefined_position : bool, optional
            If the widget should be on a special position, by default False

        Raises
        ------
        RuntimeError
            If the addition was not successful.
        """
        from .current_selection import CurrentSelectionDisplay
        super().add_widget(chemoton_widget, predefined_position)
        wrong_input = False
        previous_length = len(self._wheel_thread)
        if predefined_position:
            # special widget
            if isinstance(chemoton_widget, SelectionWidget):
                # global selection widget
                self._global_selection_widget.close()
                self._grid_layout.replaceWidget(self._global_selection_widget, chemoton_widget)
                self._global_selection_widget = chemoton_widget
                self._global_selection_widget.setMaximumWidth(300)
                self._global_selection_widget.setMinimumWidth(300)
                self._global_selection_widget.setMaximumHeight(200)
                self._global_selection_widget.setMinimumHeight(200)
                self._global_selection_widget.updateGeometry()
                self._wheel_thread.set_global_selection(chemoton_widget.exploration_step,
                                                        chemoton_widget.global_for_first)
            elif isinstance(chemoton_widget, CurrentSelectionDisplay):
                # current selection widget
                self._current_selection_widget.close()
                self._grid_layout.replaceWidget(self._current_selection_widget, chemoton_widget)
                self._current_selection_widget.setParent(None)  # type: ignore
                self._current_selection_widget = chemoton_widget
                self._current_selection_widget.updateGeometry()
            else:
                raise RuntimeError("Internal error, unknown widget")
        elif isinstance(chemoton_widget, StepWidget):
            # try to add step
            try:
                self._wheel_thread.append(chemoton_widget.exploration_step)
            except TypeError as e:
                write_error_message(f"Could not add step: {e}")
                super().delete_widget(chemoton_widget)
                wrong_input = True
        elif isinstance(chemoton_widget, SelectionWidget):
            # try to add selection
            try:
                self._wheel_thread.append(chemoton_widget.exploration_step)
            except TypeError as e:
                write_error_message(f"Could not add selection: {e}")
                super().delete_widget(chemoton_widget)
                wrong_input = True
        else:
            raise RuntimeError("Internal error, unknown widget")
        settings_index = len(self._wheel_thread) - 1
        if not predefined_position and not wrong_input and len(self._wheel_thread) == previous_length:
            # wheel combined selections, update widgets accordingly
            index = len(self.created_chemoton_widgets) - 2
            new_widget = self._construct_combined_selection_widget(index)

            new_widget.changed_settings_signal.connect(  # pylint: disable=no-member
                lambda settings: self._update_wheel_settings(settings, settings_index)
            )
            if new_widget.button_copy_results is not None:
                new_widget.button_copy_results.clicked.connect(  # pylint: disable=no-member
                    lambda: self._copy_results(new_widget)
                )
            super().add_widget(new_widget, predefined_position=False)
        elif not predefined_position and not wrong_input and len(self._wheel_thread) == previous_length + 2:
            # wheel added WholeDatabaseStep before selection
            # this is done if the first selection is not a first safe selection
            # but the database already contains structures
            added_index = len(self.created_chemoton_widgets) - 1
            added_widget = self.created_chemoton_widgets[added_index]
            assert isinstance(added_widget, SelectionWidget)
            additional_filters = added_widget.additional_filter_description
            new_selection = self._wheel_thread.get_scheme()[-1]
            assert isinstance(new_selection, Selection)
            whole_db_step = self._wheel_thread.get_scheme()[-2]
            assert isinstance(whole_db_step, GiveWholeDatabaseWithModelResult)
            super().delete_widget(added_widget)
            step_widget = StepWidget(self, whole_db_step, whole_db_step.name, whole_db_step.name)
            super().add_widget(step_widget, predefined_position=False)
            selection_widget = SelectionWidget(self, new_selection, new_selection.name, new_selection.name,
                                               additional_filter_description=additional_filters)
            super().add_widget(selection_widget, predefined_position=False)
            for i, w in enumerate([step_widget, selection_widget]):
                assert isinstance(w, ExplorationStepWidget)
                idx = len(self._wheel_thread) - i - 1
                w.changed_settings_signal.connect(  # pylint: disable=no-member
                    lambda settings, index=idx: self._update_wheel_settings(settings, index)
                )
                if w.button_copy_results is not None:
                    w.button_copy_results.clicked.connect(  # pylint: disable=no-member
                        lambda widget=w: self._copy_results(widget)
                    )
        elif isinstance(chemoton_widget, ExplorationStepWidget):
            # SteeringWheel did not mess with our addition
            # we can simply connect some signals
            chemoton_widget.changed_settings_signal.connect(  # pylint: disable=no-member
                lambda settings: self._update_wheel_settings(settings, settings_index)
            )
            if chemoton_widget.button_copy_results is not None:
                chemoton_widget.button_copy_results.clicked.connect(  # pylint: disable=no-member
                    lambda: self._copy_results(chemoton_widget)
                )

    def _construct_combined_selection_widget(self, step_index: int) -> SelectionWidget:
        """
        Deletes the two widgets that were combined and returns a new widget
        that represents the combined selection.

        Notes
        -----
        * Assumes that the combination has already been done in the SteeringWheel.
        * Assumes that exactly two widgets have been combined.

        Parameters
        ----------
        step_index : int
            The index of the new combined selection in the SteeringWheel.

        Returns
        -------
        SelectionWidget
            The new widget that represents the combined selection.
        """
        new_selection = self._wheel_thread.get_scheme()[step_index]
        assert isinstance(new_selection, Selection)
        additional_filters = ""
        for _ in range(2):
            widget = self.created_chemoton_widgets[step_index]
            assert isinstance(widget, SelectionWidget)
            if widget.additional_filter_description not in additional_filters:
                additional_filters += widget.additional_filter_description + "\n"
            super().delete_widget(widget)
        self._wheel_thread.reset_selection_index()
        return SelectionWidget(self, new_selection, new_selection.name, new_selection.name,
                               additional_filters)

    def _update_wheel_settings(self, settings: dict, index: int) -> None:
        """
        Propagates changes in the settings of a widget to the SteeringWheel.

        Parameters
        ----------
        settings : dict
            The new settings.
        index : int
            The index of the widget in the SteeringWheel.
        """
        self._wheel_thread.update_step_options(settings, index)

    def delete_widget(self, chemoton_widget: Union[StepWidget, SelectionWidget]) -> int:
        """
        Overloads the delete_widget method of the ChemotonWidgetContainer.
        Takes care that not only the widget is deleted but also corresponding step
        in the SteeringWheel.

        Parameters
        ----------
        chemoton_widget : Union[StepWidget, SelectionWidget]
            The widget to be deleted.

        Returns
        -------
        int
            The index of the deleted widget.
        """
        # avoid race condition if user presses delete button of multiple widgets,
        # before deletion is finished
        while self._is_currently_deleting:
            sleep(0.1)
        self._is_currently_deleting = True
        # take care of global selection
        if chemoton_widget is self._global_selection_widget:
            self._wheel_thread.remove_global_selection(chemoton_widget.exploration_step)
            self.add_widget(self._default_global_selection(), predefined_position=True)
            return 0
        # try to delete step + widget
        previous_length = len(self._wheel_thread)
        index = self.created_chemoton_widgets.index(chemoton_widget)
        try:
            self._wheel_thread.pop(index)
            index = super().delete_widget(chemoton_widget)
        except RuntimeError as e:
            write_error_message(str(e))
            index = -1
        else:
            # additional things if deletion was successful
            if len(self._wheel_thread) == previous_length - 2:
                # we combined some selections, update widgets accordingly
                # we assume we combined those around 'index'
                combined_widget = self._construct_combined_selection_widget(index - 1)
                widgets_after = []
                for i in range(index + 1, len(self._wheel_thread)):
                    widgets_after.append(self.created_chemoton_widgets[i])
                for widget in widgets_after:
                    super().delete_widget(widget)
                super().add_widget(combined_widget)
                for widget in widgets_after:
                    super().add_widget(widget)
        self._is_currently_deleting = False
        return index

    def _start_stop_all(self, start: bool, force_join: bool = False) -> None:
        """
        Overloads the _start_stop_all method of the ChemotonWidgetContainer.
        Simply propagates the run/stop calls with the SteeringWheel thread +
        some sanity checks to catch multiple calls etc.

        Parameters
        ----------
        start : bool
            Whether to start or stop the exploration.
        force_join : bool, optional
            Name stems from base class, if true the wheel is stopped also within a step,
            otherwise wait for the current step to finish.
        """
        if self.is_blocked:
            write_error_message("Currently in the process of stopping exploration, please be patient")
            return
        if start and force_join:
            write_error_message("Internal error, conflicting inputs")
            return
        if not self._wheel_thread:
            write_error_message("You have not specified any exploration steps, yet")
            return
        if start == self._wheel_thread.wheel_is_running():
            message = f"You wanted to {'start' if start else 'stop'} the exploration, but exploration is " \
                      f"{'already' if start else 'not'} running"
            write_error_message(message)
            return
        if start:
            self._play.setChecked(True)
            self._wheel_thread.start_wheel()
            self._play.setChecked(False)
        else:
            self._stop.setChecked(True)
            answer = False
            if self._wheel_thread and self._wheel_thread.wheel_is_running():
                answer = yes_or_no_question(
                    self, "You are currently exploring. "
                    "Do you want to wait for the current exploration step to finish (y) or "
                    "stop it within the step (n)", default_answer="no")
            self._stop_thread = Thread(target=self._stop_impl, args=(not answer,))
            self._stop_thread.start()

    def _stop_impl(self, within_step: bool) -> None:
        """
        Stops the SteeringWheel
        """
        self.is_blocked = True
        if within_step:
            self._wheel_thread.terminate_wheel(save_progress=True)
        else:
            self._wheel_thread.stop_wheel()
        self.is_blocked = False
        self._stop.setChecked(False)

    def _restart(self) -> None:
        """
        Restarts the exploration from scratch based on the given protocol
        without considering existing results.
        """
        if self.is_blocked:
            write_error_message("Currently in the process of stopping exploration, please be patient")
            return
        if not self._wheel_thread:
            write_error_message("You have not specified any exploration steps, yet")
            return
        answer = yes_or_no_question(self, "Are you sure you want to restart the exploration?")
        if not answer:
            write_info_message("Reset cancelled")
            return
        write_info_message("Stopping exploration")
        self._stop.setChecked(True)
        self._stop_impl(within_step=True)
        write_info_message("Restarting exploration")
        self._wheel_thread.start_wheel(restart=False)
        self._wheel_thread.reset_selection_index()

    def _terminate(self) -> None:
        """
        Terminates the underlying thread.
        """
        if self.is_blocked:
            write_error_message("Currently in the process of stopping exploration, please be patient")
            return
        answer = yes_or_no_question(self, "Are you sure you want to terminate the exploration")
        if not answer:
            write_info_message("Termination cancelled")
            return
        write_info_message("Terminate exploration")
        answer = yes_or_no_question(self, "Do you want to safe the progress before termination")
        self._wheel_thread.terminate_wheel(save_progress=answer)

    def construct_new_current_selection_widget(self, name: str, selection_result: SelectionResult) -> None:
        """
        Builds a new current selection widget, that shows the results the current selection gives
        for different potential expansions.

        Parameters
        ----------
        name : str
            The name of the widget
        selection_result : SelectionResult
            The result of the selection the widget should be built around

        Notes
        -----
        Since we always reconstruct the widget, this can lead to a large number of molecular widgets over
        time, which may eventually hit the 100 mark.
        This can be avoided by a redesign by separating the molecular widget, or update the whole widget
        instead of rebuilding it every time.
        """
        from .current_selection import CurrentSelectionDisplay
        # push quickly constructed loading info widget
        self.add_widget(CurrentSelectionDisplay.with_loading_info(self), predefined_position=True)
        # now construct real thing which requires some querying
        selection_display = CurrentSelectionDisplay(self, self.db_manager, selection_result, name)
        self.add_widget(selection_display, predefined_position=True)

    def color_widget(self, index: int, status: Status) -> None:
        """
        Colors a step widget based on its status.

        Parameters
        ----------
        index : int
            Index of the widget
        status : Status
            The status of the widget
        """
        if status == Status.FINISHED:
            color = StartStopWidget.green
        elif status == Status.FAILED:
            color = StartStopWidget.red
        elif status == Status.CALCULATING:
            color = StartStopWidget.orange
        else:
            color = ""
        self.created_chemoton_widgets[index].change_color(color)

    def activate_copy(self, index: int, status: Status) -> None:
        """
        Enables the copy results button of a widget if its type and status are appropriate.

        Parameters
        ----------
        index : int
            The index of the widget
        status : Status
            The status of the widget
        """
        if status != Status.FINISHED:
            return
        widget = self.created_chemoton_widgets[index]
        if isinstance(widget, SelectionWidget) and widget.button_copy_results is not None:
            widget.button_copy_results.setEnabled(True)

    def _copy_results(self, widget: SelectionWidget) -> None:
        """
        Creates and adds a new widget to our protocol that gives the same result as the given
        selection widget.

        Parameters
        ----------
        widget : SelectionWidget
            The widget we copy the results from
        """
        if not isinstance(widget, SelectionWidget):
            write_error_message("Failed to copy results: Exploration step is not a selection")
            return
        try:
            index = self.created_chemoton_widgets.index(widget)
        except ValueError:
            write_error_message("Failed to copy results: Selection not found")
            return
        results = self._wheel_thread.get_results()
        if results is None or index >= len(results):
            write_error_message("Failed to copy results: Results not found")
            return
        result = results[index]
        if not isinstance(result, SelectionResult):
            write_error_message("Failed to copy results: Results are not of type SelectionResult")
            return
        self._pre_io_operations()  # necessary because this ensures that selection in widget has results
        selection = widget.exploration_step
        new_widget = SelectionWidget(
            self,
            PredeterminedSelection(selection.options.model, result, logic_coupling=selection.logic_coupling),
            f"Copied results from {widget.label} at index {index}",
            widget.name, widget.additional_filter_description, False, True, False
        )
        self.add_widget(new_widget, predefined_position=False)

    def _generate_save_object(self) -> List[List[Any]]:
        """
        Puts all current widgets into one object from which the widgets can be reconstructed.

        Returns
        -------
        List[List[Any]]
            The save object
        """
        save_object = super()._generate_save_object()
        save_object.append([self._wheel_thread.get_partial_restart_info()])
        return save_object

    def _unpack_save_object(self, save_object: List[List[Any]]) -> None:
        """
        Reconstructs all widgets from the given save object.

        Parameters
        ----------
        save_object : List[List[Any]]
            The save object
        """
        partial_steps = save_object.pop(-1)[0]
        # for backwards compatibility
        if not isinstance(partial_steps, RestartPartialExpansionInfo) and isinstance(partial_steps, Model):
            save_object.append([partial_steps])
            partial_steps = None
        super()._unpack_save_object(save_object)
        self._wheel_thread.set_partial_restart_info(partial_steps)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
"""
Provides the ElectronicDataStatusManager class.
"""

from typing import Optional, TYPE_CHECKING, Any
from PySide2.QtCore import QObject
if TYPE_CHECKING:
    Signal = Any
else:
    from PySide2.QtCore import Signal


class ElectronicDataStatusManager(QObject):
    """
    Provides a Status Manager that hold a list and emits events on resize
    """

    molden_input_changed_signal = Signal(str)
    hamiltonian_changed = Signal()

    def __init__(self, parent: Optional[QObject] = None):
        super().__init__(parent)
        self.__molden_input = str()

    @property
    def molden_input(self) -> str:
        """
        Returns the error_message.
        """
        return self.__molden_input

    @molden_input.setter
    def molden_input(self, value: str) -> None:
        """
        Sets the contained value. Notifies on change.
        """
        if value is None or value == "":
            return

        if value == self.__molden_input:
            return

        self.__molden_input = value
        self.molden_input_changed_signal.emit(value)

    def signal_hamiltonian_change(self) -> None:
        self.hamiltonian_changed.emit()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

from typing import List, Optional

from PySide2.QtWidgets import QVBoxLayout, QHBoxLayout, QWidget, QLayout


class VerticalLayout(QVBoxLayout):
    """
    Inherits from QVBoxLayout and is equal in all aspects but can be constructed
    with multiple widgets to be in the layout add once and can add multiple widgets at
    once later.
    """

    def __init__(self, widgets: Optional[List[QWidget]] = None):
        """
        Construct the widget with an optional list of widgets to be added to the
        constructed layout. The order in the list determines the order in the layout.

        Parameters
        ----------
        widgets : Optional[List[QWidget]], optional
            The optional widgets for the layout, by default None
        """
        super().__init__()
        if widgets is not None:
            for w in widgets:
                self.addWidget(w)

    def add_widgets(self, widgets: List[QWidget]) -> None:
        """
        Add multiple widgets to the layout.
        The order in the list determines the order in the layout.

        Parameters
        ----------
        widgets : List[QWidget]
            The added widgets
        """
        for w in widgets:
            self.addWidget(w)

    def add_layouts(self, layouts: List[QLayout]) -> None:
        """
        Add multiple sublayouts to the layout.
        The order in the list determines the order in the layout.

        Parameters
        ----------
        layouts : List[QLayout]
            The added layouts
        """
        for layout in layouts:
            self.addLayout(layout)


class HorizontalLayout(QHBoxLayout):
    """
    Inherits from QHBoxLayout and is equal in all aspects but can be constructed
    with multiple widgets to be in the layout add once and can add multiple widgets at
    once later.
    """

    def __init__(self, widgets: Optional[List[QWidget]] = None):
        """
        Construct the widget with an optional list of widgets to be added to the
        constructed layout. The order in the list determines the order in the layout.

        Parameters
        ----------
        widgets : Optional[List[QWidget]], optional
            The optional widgets for the layout, by default None
        """
        super().__init__()
        if widgets is not None:
            for w in widgets:
                self.addWidget(w)

    def add_widgets(self, widgets: List[QWidget]) -> None:
        """
        Add multiple widgets to the layout.
        The order in the list determines the order in the layout.

        Parameters
        ----------
        widgets : List[QWidget]
            The added widgets
        """
        for w in widgets:
            self.addWidget(w)

    def add_layouts(self, layouts: List[QLayout]) -> None:
        """
        Add multiple sublayouts to the layout.
        The order in the list determines the order in the layout.

        Parameters
        ----------
        layouts : List[QLayout]
            The added layouts
        """
        for layout in layouts:
            self.addLayout(layout)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
"""
Provides the ElectronicDataImageGenerator class.
"""

import math
import numpy as np
from typing import List, Dict
import scine_utilities as su
from vtk import (
    vtkImageData,
    VTK_DOUBLE,
)
from scine_heron.electronic_data.electronic_data import ElectronicData


class ElectronicDataImageGenerator:
    """
    Generate vtkImageData that contains electronic data.
    Important: Code here only works in spherical harmonics and not in Cartesians.
    """

    def __init__(self, electronic_data: ElectronicData):
        self.__electronic_data = electronic_data
        self.__block_size = 8
        self.__threshold = 1e-2
        self.__step_size = 0.2
        self.__dim = self.__get_box_size()
        self.__cache: Dict[int, np.ndarray] = dict()
        self.__steps = [
            int((self.__dim[1] - self.__dim[0]) / self.__step_size + 0.5),
            int((self.__dim[3] - self.__dim[2]) / self.__step_size + 0.5),
            int((self.__dim[5] - self.__dim[4]) / self.__step_size + 0.5),
        ]
        self.__dx, self.__dy, self.__dz = (
            (self.__dim[1] - self.__dim[0]) / (self.__steps[0] - 1),
            (self.__dim[3] - self.__dim[2]) / (self.__steps[1] - 1),
            (self.__dim[5] - self.__dim[4]) / (self.__steps[2] - 1),
        )

    def __get_box_size(self) -> List[float]:
        x = np.array(
            [
                self.__electronic_data.atoms[index].coordinates[0]
                for index in range(len(self.__electronic_data.atoms))
            ]
        )
        y = np.array(
            [
                self.__electronic_data.atoms[index].coordinates[1]
                for index in range(len(self.__electronic_data.atoms))
            ]
        )
        z = np.array(
            [
                self.__electronic_data.atoms[index].coordinates[2]
                for index in range(len(self.__electronic_data.atoms))
            ]
        )
        borders = np.array(
            [
                np.sqrt(-np.log(self.__threshold * 10) / atom.min_alpha)
                for atom in self.__electronic_data.atoms
            ]
        )
        box = [
            min(x - borders),
            max(x + borders),
            min(y - borders),
            max(y + borders),
            min(z - borders),
            max(z + borders),
        ]
        return box

    def generate_mo_image(self, orbital_index: int) -> vtkImageData:
        """
        Generate molecular orbital image.
        """
        image = vtkImageData()
        image.SetDimensions(self.__steps[0], self.__steps[1], self.__steps[2])
        image.SetOrigin(self.__dim[0], self.__dim[2], self.__dim[4])
        image.SetSpacing(self.__dx, self.__dy, self.__dz)
        image.AllocateScalars(VTK_DOUBLE, 1)
        image.GetPointData().GetScalars().Fill(0.0)

        block_index = 0
        i = 0
        x = self.__dim[0]
        while i < self.__steps[0]:
            j = 0
            y = self.__dim[2]
            while j < self.__steps[1]:
                k = 0
                z = self.__dim[4]
                while k < self.__steps[2]:
                    block = None
                    if orbital_index == -4:
                        # Electron Density
                        for orb in self.__electronic_data.mo:
                            coefficients_size = len(orb.coefficients)
                            if block_index not in self.__cache:
                                self.__calc_mo_at_block_and_save_in_cache(
                                    block_index, coefficients_size, x, y, z
                                )
                            block_tmp = self.__cache[block_index].dot(orb.coefficients)
                            block_tmp = np.square(block_tmp)
                            try:
                                block += block_tmp
                            except BaseException:
                                block = block_tmp
                    else:
                        mo = self.__electronic_data.mo[orbital_index]
                        coefficients_size = len(mo.coefficients)
                        if block_index not in self.__cache:
                            self.__calc_mo_at_block_and_save_in_cache(
                                block_index, coefficients_size, x, y, z
                            )
                        block = self.__cache[block_index].dot(mo.coefficients)
                    if block is not None and np.sum(np.abs(block)) > 1.0e-4:
                        self.__add_mo_block_at_image(image, block, i, j, k)

                    k += self.__block_size
                    z += self.__dz * self.__block_size
                    block_index += 1
                j += self.__block_size
                y += self.__dy * self.__block_size
            i += self.__block_size
            x += self.__dx * self.__block_size
        return image

    def __add_mo_block_at_image(
        self,
        image: vtkImageData,
        block: np.ndarray,
        origin_i: int,
        origin_j: int,
        origin_k: int,
    ) -> None:
        index = 0
        for i in range(self.__block_size):
            for j in range(self.__block_size):
                for k in range(self.__block_size):
                    if (
                        origin_i + i < self.__steps[0]
                        and origin_j + j < self.__steps[1]
                        and origin_k + k < self.__steps[2]
                    ):
                        if block[index] != 0.0:
                            image.SetScalarComponentFromDouble(
                                origin_i + i,
                                origin_j + j,
                                origin_k + k,
                                0,
                                block[index],
                            )
                        index += 1

    def __nearest_value_in_list(
        self, atom_value: float, block_value: float, step: float
    ) -> float:
        max_block_val = block_value + (self.__block_size - 1) * step
        if atom_value < block_value:
            return block_value
        elif atom_value > max_block_val:
            return max_block_val
        else:
            return block_value + int(((atom_value - block_value) / step) + 0.5) * step

    def __calc_mo_at_block_and_save_in_cache(
        self, block_index: int, coefficients_size: int, x: float, y: float, z: float,
    ) -> None:
        """
        Calculate the MO value at given point.
        """
        chi_index = 0
        self.__cache[block_index] = np.zeros(
            (
                self.__block_size * self.__block_size * self.__block_size,
                coefficients_size,
            )
        )
        block = self.__cache[block_index].reshape(
            (
                self.__block_size
                * self.__block_size
                * self.__block_size
                * coefficients_size
            )
        )

        for atom in self.__electronic_data.atoms:
            atom_x = atom.coordinates[0]
            atom_y = atom.coordinates[1]
            atom_z = atom.coordinates[2]

            nearest_x = self.__nearest_value_in_list(atom_x, x, self.__dx)
            nearest_y = self.__nearest_value_in_list(atom_y, y, self.__dy)
            nearest_z = self.__nearest_value_in_list(atom_z, z, self.__dz)

            nearest_x_ang = (nearest_x - atom_x) * su.BOHR_PER_ANGSTROM
            nearest_y_ang = (nearest_y - atom_y) * su.BOHR_PER_ANGSTROM
            nearest_z_ang = (nearest_z - atom_z) * su.BOHR_PER_ANGSTROM

            ra2_nearest = (
                nearest_x_ang * nearest_x_ang
                + nearest_y_ang * nearest_y_ang
                + nearest_z_ang * nearest_z_ang
            )

            if ra2_nearest > 1:
                if math.exp(-ra2_nearest * atom.min_alpha) < self.__threshold:
                    chi_index += atom.sum_chi_step
                    continue

            for orbital in atom.gaussian_orbitals:
                if ra2_nearest > 1:
                    if (
                        np.dot(orbital.coeff, np.exp(-ra2_nearest * orbital.alpha))
                        < self.__threshold
                    ):
                        chi_index += orbital.chi_step()
                        continue

                for i in range(self.__block_size):
                    xx = x + (i * self.__dx)
                    xa = (xx - atom_x) * su.BOHR_PER_ANGSTROM
                    xa2 = xa * xa
                    iOff = i * self.__block_size

                    for j in range(self.__block_size):
                        yy = y + (j * self.__dy)
                        ya = (yy - atom_y) * su.BOHR_PER_ANGSTROM
                        ya2 = ya * ya
                        ijOff = (iOff + j) * self.__block_size

                        for k in range(self.__block_size):
                            zz = z + (k * self.__dz)
                            za = (zz - atom_z) * su.BOHR_PER_ANGSTROM
                            za2 = za * za
                            ijkOff = (ijOff + k) * coefficients_size + chi_index

                            ra2 = xa2 + ya2 + za2

                            radial_sum = np.dot(
                                orbital.coeff, np.exp(-ra2 * orbital.alpha)
                            )

                            if radial_sum < self.__threshold:
                                continue

                            if orbital.orb_type == "s":  # s orbital
                                block[ijkOff] = radial_sum
                            elif orbital.orb_type == "p":  # p orbital
                                block[ijkOff] = xa * radial_sum
                                block[ijkOff + 1] = ya * radial_sum
                                block[ijkOff + 2] = za * radial_sum
                            elif orbital.orb_type == "d":  # d orbital [5D]
                                block[ijkOff] = (
                                    0.288675135 * (2 * za2 - xa2 - ya2) * radial_sum
                                )
                                block[ijkOff + 1] = 0.5 * (xa2 - ya2) * radial_sum
                                block[ijkOff + 2] = xa * ya * radial_sum
                                block[ijkOff + 3] = xa * za * radial_sum
                                block[ijkOff + 4] = ya * za * radial_sum
                            elif orbital.orb_type == "f":  # f orbital [7F]
                                block[ijkOff] = (
                                    radial_sum * za * (5.0 * za2 - 3.0 * ra2)
                                )
                                block[ijkOff + 1] = radial_sum * xa * (5.0 * za2 - ra2)
                                block[ijkOff + 2] = radial_sum * ya * (5.0 * za2 - ra2)
                                block[ijkOff + 3] = radial_sum * za * (xa2 - ya2)
                                block[ijkOff + 4] = radial_sum * xa * ya * za
                                block[ijkOff + 5] = radial_sum * (
                                    xa * xa2 - 3.0 * xa * ya2
                                )
                                block[ijkOff + 6] = radial_sum * (
                                    3.0 * xa2 * ya - ya2 * ya
                                )
                            else:
                                raise NotImplementedError(
                                    "The atomic orbital type '"
                                    + orbital.orb_type
                                    + "' is not implemented."
                                )
                chi_index += orbital.chi_step()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
"""
Provides the HapticClient class.
"""
from PySide2.QtCore import QObject
import typing
if typing.TYPE_CHECKING:
    Signal = typing.Any
else:
    from PySide2.QtCore import Signal

try:
    import scine_heron_haptic as suh
except ImportError:
    pass


class HapticSignals(QObject):
    """
    Inherited from QObject.
    Connect callback to UI.
    """

    move_signal = Signal(object, float, float, float)
    first_button_down_signal = Signal()
    first_button_up_signal = Signal()

    second_button_down_signal = Signal(object)
    second_button_up_signal = Signal()


class HapticCallback(suh.HapticCallback):  # type: ignore[misc]
    """
    Inherited from HapticCallback.
    Connect to haptic events.
    """

    def __init__(self) -> None:
        super().__init__()
        self.signals = HapticSignals()

    def move(
        self, data: suh.HapticData, azimuth: float, elevation: float, zoom: float
    ) -> None:
        self.signals.move_signal.emit(data, azimuth, elevation, zoom)

    def first_button_down(self) -> None:
        self.signals.first_button_down_signal.emit()

    def first_button_up(self) -> None:
        self.signals.first_button_up_signal.emit()

    def second_button_down(self, atom_index: int) -> None:
        self.signals.second_button_down_signal.emit(atom_index)

    def second_button_up(self) -> None:
        self.signals.second_button_up_signal.emit()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
from random import random, randint
from time import sleep
from typing import Dict, Union, Optional, Any, TYPE_CHECKING, List, Tuple, Callable
from threading import Thread

from PySide2.QtGui import QCloseEvent
from PySide2.QtCore import QThread, Qt
from PySide2.QtWidgets import (
    QWidget,
    QHBoxLayout,
    QLabel,
    QLineEdit,
)
from PySide2.QtCore import QObject
if TYPE_CHECKING:
    Signal = Any
else:
    from PySide2.QtCore import Signal

from scine_chemoton.gears import Gear
from scine_chemoton.gears.elementary_steps import ElementaryStepGear
from scine_chemoton.engine import Engine  # pylint: disable=import-error
from scine_chemoton.steering_wheel.network_expansions import NetworkExpansion  # pylint: disable=import-error
from scine_chemoton.steering_wheel.selections import Selection  # pylint: disable=import-error
from scine_chemoton.steering_wheel.datastructures import GearOptions  # pylint: disable=import-error

from scine_heron.containers.start_stop_widget import StartStopWidget
from scine_heron.containers.buttons import TextPushButton
from scine_heron.containers.layouts import HorizontalLayout
from scine_heron.containers.wrapped_label import WrappedLabel
from scine_heron.chemoton.class_searcher import ChemotonClassSearcher
from scine_heron.settings.class_options_widget import ClassOptionsWidget
from scine_heron.settings.dict_option_widget import DictOptionWidget
from scine_heron.settings.docstring_parser import DocStringParser
from scine_heron.utilities import write_info_message, write_error_message, construct_sound


class EngineWidget(StartStopWidget):
    """
    A widget that holds a single Chemoton engine, its gear, and its settings.
    It allows interaction with them via buttons.
    """
    # signal that is emitted when a single run is finished
    # this exists, so that this widget's thread class can communicate with the widget.
    single_run_finished = Signal()

    class SingleRunChecker(QThread):
        """
        A utility class for the engine widget that checks whether a single run is finished.
        This is useful, because then the widget can join the engine.
        """

        def __init__(self, parent: Optional[QWidget], engine: Engine, finished_signal: Signal):
            """
            Construct the Thread.
            Every instance of this class can only work on a single engine.

            Parameters
            ----------
            parent : Optional[QWidget]
                The parent widget.
            engine : Engine
                The engine we should observe
            finished_signal : Signal
                The signal we emit, when the engine has gone through a single run.
            """
            super().__init__(parent)
            self._counter = 0
            self._signal = finished_signal
            self._engine = engine
            self._stop = False
            self._sleep_time = 0.1

        def is_running(self) -> bool:
            """
            Whether this thread is running
            """
            return not self._stop

        def stop(self) -> None:
            """
            Stop the thread (not the engine!)
            """
            self._stop = True
            sleep(2 * self._sleep_time)

        def run(self):
            """
            The main loop of the thread.
            It checks the number of gear loops and emits a signal, when it has changed and then exits.
            """
            self._stop = False
            new_counter = self._counter
            while new_counter == self._counter and not self._stop:
                sleep(self._sleep_time)
                new_counter = self._engine.get_number_of_gear_loops()
            self._signal.emit()
            self.exit(0)

    def __init__(
        self,
        parent: Optional[QObject],
        engine: Engine,
        gear: Gear,
        engine_label: str,
        gear_name: str,
        filter_description: str = "None",
    ) -> None:
        """
        Construct the engine widget with the engine, its gear and additional descriptions.

        Parameters
        ----------
        parent : Optional[QObject]
            The parent widget.
        engine : Engine
            The engine to be run
        gear : Gear
            The gear of the engine
        engine_label : str
            The label of this widget
        gear_name : str
            The name of the gear
        filter_description : str, optional
            A description of the set filters, by default "None"
        """
        super().__init__(parent)
        self._parent = parent

        self.setMaximumWidth(300)

        self.engine = engine
        self.gear = gear
        if isinstance(gear, ElementaryStepGear):
            self._sound_start = construct_sound("engine-2")
        else:
            self._sound_start = construct_sound(f"engine-{randint(0, 1)}")
        self._sound_stop = construct_sound("break")
        self._sound_stop.stop()
        self._sound_start.stop()
        self.__engine_is_working = False
        self.__engine_is_joining = False

        # Create layout and add widgets
        self._layout.addWidget(WrappedLabel(engine_label))

        self.__add_gear_at_layout(gear_name)
        self._add_filter_at_layout(filter_description)

        self.button_start_stop = TextPushButton("Start", self.start_stop)
        self.button_single = TextPushButton("Single Run", self.__single_engine_run)

        self._layout.addWidget(self.button_start_stop)
        self._layout.addWidget(self.button_single)
        self.single_run_finished.connect(self.start_stop)  # pylint: disable=no-member
        self.__single_observer: Optional[Any] = None
        self._join_thread: Optional[Thread] = None

        self._add_settings_and_delete_buttons()

        self.init_arguments += [self.engine, self.gear, engine_label, gear_name, filter_description]

        self.setLayout(self._layout)

    def closeEvent(self, event: QCloseEvent) -> None:
        """
        Clean up our engine before we close the widget.

        Parameters
        ----------
        event : QCloseEvent
            The close event.
        """
        if self._join_thread is not None:
            self._join_thread.join()
        if self.__single_observer is not None and self.__single_observer.is_running():
            self.__single_observer.stop()
            self.__single_observer.wait()
        super().closeEvent(event)

    def set_docstring_dict(self, doc_string_parser: DocStringParser) -> None:
        """
        Set the docstring dictionary for this widget based on our gear and if it is an elementary step gear,
        the trial generator of the gear.

        Parameters
        ----------
        doc_string_parser : DocStringParser
            The parser that can extract the docstrings.
        """
        self._docstring_dict = doc_string_parser.get_docstring_for_object_attrs(
            self.gear.__class__.__name__, self.gear.options
        )
        if isinstance(self.gear, ElementaryStepGear):
            # additional optional stuff
            additionals: List[Dict[str, str]] = []
            if hasattr(self.gear, "trial_generator"):
                obj = self.gear.trial_generator.options
                additionals.append(doc_string_parser.get_docstring_for_object_attrs(obj.__class__.__name__, obj))
            for add in additionals:
                self._docstring_dict = {**self._docstring_dict, **add}

    def stop_class_if_working(self) -> None:
        """
        Stop our engine.
        """
        if self.__engine_is_working:
            if self.__single_observer is not None and self.__single_observer.is_running():
                self.__single_observer.stop()
            self.engine.stop()

    def is_running(self) -> bool:
        return self.__engine_is_working

    def __single_engine_run(self) -> None:
        """
        Run the engine a single time, unless it is already running.
        """
        if self.__engine_is_working:
            write_info_message("Engine is still running")
            return
        self.__switch_status()
        self.__single_observer = self.SingleRunChecker(self, self.engine, self.single_run_finished)
        self.__single_observer.start()
        self.engine.run(single=True)
        if self.button_settings is not None:
            self.button_settings.setEnabled(False)
        if self.button_delete is not None:
            self.button_delete.setEnabled(False)
        # make sure run is not stopped at break point
        self.gear.stop_at_break_point(False)
        self.engine.set_gear(self.gear)

    def __switch_status(self):
        """
        Switch our own status.
        """
        if self.__engine_is_working:
            self.change_color()
            self.button_start_stop.setText("Start")
            if getattr(self._parent, "sound_allowed", False):
                sleep(random() * 0.5)
                self._sound_start.stop()
                self._sound_stop.play()
        else:
            self.change_color(self.green)
            self.button_start_stop.setText("Stop")
            if getattr(self._parent, "sound_allowed", False):
                sleep(random() * 0.5)
                self._sound_stop.stop()
                self._sound_start.play()
        self.__engine_is_working = not self.__engine_is_working

    def start_stop(self, start_all: bool = False, stop_all: bool = False) -> None:
        """
        Start the engine if it is not running, stop it otherwise.

        Parameters
        ----------
        start_all : bool, optional
            Ensures that our engine is not stopped, if it is already running if set to True, by default False
        stop_all : bool, optional
            Influences the joining behavior, if set to True, because if multiple engines
            are stopped, the joining can be simplified, by default False
        """
        if start_all and stop_all:
            write_error_message("Internal Error")
            return
        if start_all and not self.__engine_is_working:
            self._run_impl()
        elif stop_all and self.__engine_is_working:
            if self.__single_observer is not None and self.__single_observer.is_running():
                self.__single_observer.stop()
            if self.__engine_is_joining:
                # expect separate join() call from caller of this method with stop_all=True
                return
            self.engine.stop()
        elif not start_all and self.__engine_is_working:
            if self.__single_observer is not None and self.__single_observer.is_running():
                self.__single_observer.stop()
            if self.__engine_is_joining:
                write_info_message("Still waiting for previous loop of engine to finish")
                return
            self.engine.stop()
            # do join with thread here
            self._join_thread = Thread(target=self.join)
            self._join_thread.start()
        elif not stop_all and not self.__engine_is_working:
            self._run_impl()

    def _run_impl(self) -> None:
        """
        Run the engine, ensure we have the right status and control settings access.
        """
        self.engine.run()
        self.__switch_status()
        if self.button_settings is not None:
            self.button_settings.setEnabled(False)
        if self.button_delete is not None:
            self.button_delete.setEnabled(False)

    def join(self, force_join: bool = False) -> None:
        """
        Join the engine if we are working or want to enforce join.

        Parameters
        ----------
        force_join : bool, optional
            Join the engine even if we are not working, by default False
        """
        if not self.__engine_is_working and not force_join:
            # no join required if not working
            return
        if self.__engine_is_joining:
            while self.__engine_is_joining:
                # in case joining was initiated before and is in progress, we let the current join finish
                # and don't run another join, but this method is expected to only return once it was joined,
                # so, we imitate join behavior by stalling here
                sleep(0.1)
            # join is concluded, simply return
            return
        self.__engine_is_joining = True
        self.engine.join()
        if not force_join or self.__engine_is_working:
            # change if we don't force join, or if we are force joining a running engine
            self.__switch_status()
        if self.button_settings is not None:
            self.button_settings.setEnabled(True)
        if self.button_delete is not None:
            self.button_delete.setEnabled(True)
        # make sure next run is not stopped at break point
        self.gear.stop_at_break_point(False)
        self.engine.set_gear(self.gear)
        self.__engine_is_joining = False

    def __add_gear_at_layout(self, gear_name: str) -> None:
        """
        Adds the information about the gear to our layout

        Parameters
        ----------
        gear_name : str
            The name of the gear
        """
        layout = HorizontalLayout()
        layout.setAlignment(Qt.AlignLeft)

        self.gear_label = QLabel("Gear:")
        self.gear_widget = WrappedLabel(gear_name)

        layout.add_widgets([self.gear_label, self.gear_widget])

        self._layout.addLayout(layout)

    @staticmethod
    def get_settings(gear: Gear) -> Tuple[Dict[str, Any], Dict[str, Dict[str, Any]]]:
        """
        The EngineWidget holds a getter/setter for the settings, because they are more
        complicated to extract due to varying APIs between the different gears.

        Notes
        -----
        The getter/setter are static methods so that our classes can access the settings of some gears.

        Parameters
        ----------
        gear : Gear
            The gear we want the settings from

        Returns
        -------
        Tuple[Dict[str, Any], Dict[str, Dict[str, Any]]]
            The settings and an optional mapping that allows to map the settings back to the object
        """
        if not isinstance(gear, ElementaryStepGear):
            return DictOptionWidget.get_attributes_of_object(gear.options), {}
        # gather some more options
        option_attrs = DictOptionWidget.get_attributes_of_object(gear.options)
        trial_attrs = DictOptionWidget.get_attributes_of_object(gear.trial_generator.options)
        final_trial_attrs: Dict[str, Any] = {}
        attr_mapping = {}
        for k, v in trial_attrs.items():
            if k.endswith("options"):
                additional_attr = DictOptionWidget.get_attributes_of_object(v)
                # ensure unique keys
                mod_additional_attr = {f"{k}.{subkey}": subvalue for subkey, subvalue in additional_attr.items()}
                attr_mapping[k] = mod_additional_attr
                final_trial_attrs = {**final_trial_attrs, **mod_additional_attr}
            else:
                final_trial_attrs[k] = v
        return {**option_attrs, **final_trial_attrs}, attr_mapping

    @staticmethod
    def set_settings(gear: Gear, options: Dict[str, Any], attribute_mapping: Dict[str, Dict[str, Any]]) -> None:
        """
        The EngineWidget holds a getter/setter for the settings, because they are more
        complicated to extract due to varying APIs between the different gears.

        Notes
        -----
        The getter/setter are static methods so that our classes can access the settings of some gears.

        Parameters
        ----------
        gear : Gear
            The gear we want the settings from
        options : Dict[str, Any]
            The settings we want to set
        attribute_mapping : Dict[str, Dict[str, Any]]
            The mapping of the settings to the actual object
        """
        if not isinstance(gear, ElementaryStepGear):
            # easy case
            DictOptionWidget.set_attributes_to_object(gear.options, options)
            return
        # gather partial options
        option_attrs = DictOptionWidget.get_attributes_of_object(gear.options)
        trial_attrs = DictOptionWidget.get_attributes_of_object(gear.trial_generator.options)
        # apply options
        for k, v in options.items():
            try:
                if k in option_attrs:
                    setattr(gear.options, k, v)
                if k in trial_attrs:
                    setattr(gear.trial_generator.options, k, v)
                if "." in k:
                    mapping_key = k.split(".")[0]
                    if mapping_key in attribute_mapping:
                        # TODO previously relied on mapping value here, but GearOptions in NetworkExpansion
                        # made it necessary to map back to the original value, so the mapping
                        # datastructure could actually be simplified to a single dictionary
                        for subkey in attribute_mapping[mapping_key].keys():
                            unmod_key = subkey.split(".")[-1]
                            if unmod_key == k:
                                setattr(getattr(gear.trial_generator.options, mapping_key), unmod_key, v)
                                break
            except NotImplementedError:
                pass  # can be caused by trial generator warning that settings are ignored, but doesn't matter here

    def _show_settings(self) -> None:
        """
        Display the gear settings in a pop that allows to edit the settings.
        """
        options, mapping = self.get_settings(self.gear)
        setting_widget = ClassOptionsWidget(
            options, self._docstring_dict, parent=self, add_close_button=True, allow_removal=False
        )
        setting_widget.exec_()
        self.set_settings(self.gear, options, mapping)
        self.engine.set_gear(self.gear)


class ExplorationStepWidget(StartStopWidget):
    """
    An abstract base class for widget in a SteeringWheel exploration protocol.
    Can combine identical methods our information access from holding widgets.
    """
    exploration_step_label = "ExplorationStep:"
    changed_settings_signal = Signal(dict)

    def __init__(self, parent: Optional[QObject], exploration_step: Union[Selection, NetworkExpansion],
                 label: str, name: str, *args, **kwargs) -> None:
        """
        Construct the widget with the underlying exploration step and some name information.

        Parameters
        ----------
        parent : Optional[QObject]
            The parent widget
        exploration_step : Union[Selection, NetworkExpansion]
            The held exploration step that should be run
        label : str
            The label of the widget
        name : str
            The name of the step
        """
        super().__init__(parent)
        self.exploration_step = exploration_step
        self.label = label
        self.name = name

        self._layout.addWidget(WrappedLabel(label))
        self.chemical_step_widget: Optional[QLineEdit] = None
        self.init_arguments += [self.exploration_step, self.label, self.name]

    def _create_step_sub_layout(self):
        """
        Add the information about the exploration step to our widget.
        """
        sub_layout = QHBoxLayout()
        sub_layout.setAlignment(Qt.AlignLeft)
        sub_layout.addWidget(WrappedLabel(self.exploration_step_label))

        self.chemical_step_widget = WrappedLabel(self.name)
        sub_layout.addWidget(self.chemical_step_widget)

        self._layout.addLayout(sub_layout)

    def set_docstring_dict(self, doc_string_parser: DocStringParser) -> None:
        """
        Set our docstring information based on a parser that can extract it from the step.

        Parameters
        ----------
        doc_string_parser : DocStringParser
            The parser to extract with
        """
        self._docstring_dict = doc_string_parser.get_docstring_for_instance_init(
            self.exploration_step.__class__.__name__, self.exploration_step.options
        )

    def stop_class_if_working(self) -> None:
        """
        To be implemented by child class
        """

    def start_stop(self, start_all: bool = False, stop_all: bool = False) -> None:
        """
        To be implemented by child class
        """

    def join(self, force_join: bool = False) -> None:
        """
        To be implemented by child class
        """


class SelectionWidget(ExplorationStepWidget):
    """
    The widget that holds a Selection in a steering exploration protocol.
    """
    exploration_step_label_text = "Selection:"

    def __init__(self, parent: QObject, selection: Selection, label: str, selection_name: str,
                 additional_filter_description: str = "None", as_global: bool = False,
                 with_delete: bool = True, global_for_first: bool = False) -> None:
        """
        Constructs the widget with the selection, some name information, and if it is a global selection.

        Parameters
        ----------
        parent : QObject
            The parent widget
        selection : Selection
            The underlying selection
        label : str
            The name of this widget
        selection_name : str
            The name of the selection
        additional_filter_description : str, optional
            Description of any additional filters, by default "None"
        as_global : bool, optional
            If this selection is a global selection, by default False
        with_delete : bool, optional
            If a delete button should be added to our layout, by default True
        global_for_first : bool, optional
            If this global selection should be applied to the first step
            in an exploration protocol, by default False

        Raises
        ------
        RuntimeError
            If conflicting arguments concerning the global selection are given.
        """
        if global_for_first and not as_global:
            raise RuntimeError("Internal error")
        if as_global:
            this_label = f"Global: {label}"
        else:
            this_label = label
        super().__init__(parent, selection, this_label, selection_name)
        self._parent = parent
        self.additional_filter_description = additional_filter_description
        self.as_global = as_global
        self.global_for_first = global_for_first
        self._create_step_sub_layout()
        self._add_filter_at_layout(additional_filter_description)

        if not as_global:
            self.button_copy_results = TextPushButton("Copy results as last selection")
            self._layout.addWidget(self.button_copy_results)
            self.button_copy_results.setEnabled(False)  # can only be enabled after the first run
        if with_delete:
            self._add_settings_and_delete_buttons()

        self.init_arguments += [self.additional_filter_description, self.as_global, with_delete, self.global_for_first]

        self.setLayout(self._layout)

    def _show_settings(self) -> None:
        """
        Display the settings in a pop-up and manipulate them based on changes.
        """
        setting_widget = ClassOptionsWidget(
            self.exploration_step.options.__dict__, self._docstring_dict, parent=self, add_close_button=True,
            allow_removal=False
        )
        setting_widget.exec_()
        self.changed_settings_signal.emit(self.exploration_step.options.__dict__)


class StepWidget(ExplorationStepWidget):
    """
    Constructs a widget holding a NetworkExpansion.
    """
    exploration_step_label_text = "Network Expansion:"

    def __init__(self, parent: Optional[QObject], step: NetworkExpansion, label: str, step_name: str,
                 gear_options: Optional[GearOptions] = None) -> None:
        """
        Construct the widget with the Expansion, some name information and the gear options of the step.

        Parameters
        ----------
        parent : Optional[QObject]
            The parent widget
        step : NetworkExpansion
            The underlying network expansion.
        label : str
            The name of the widget
        step_name : str
            The name of the network expansion
        gear_options : Optional[GearOptions], optional
            The optional options of all the gears of the network expansion, by default None
        """
        super().__init__(parent, step, label, step_name)
        self.gear_options = gear_options
        self._create_step_sub_layout()
        self._add_settings_and_delete_buttons()
        self.setLayout(self._layout)
        self.init_arguments.append(self.gear_options)

    def _show_settings(self) -> None:
        """
        Displays the settings of the network expansion including its gear options in a pop-up
        and manipulate them based on changes.

        Notes
        -----
        Accesses main window to get the current database credentials so that we can set-up the
        network expansion without running it to get the correct gear options.
        """
        from scine_heron import find_main_window
        main = find_main_window()
        suggestions: Optional[dict] = None
        if main is not None:
            credentials = main.toolbar.current_credentials()
            if credentials is not None:
                self.exploration_step.dry_setup_protocol(credentials)
                current_gears = self.exploration_step.current_gears()
                suggestions = generate_gear_name_settings_suggestions(current_gears)
                # very important to clean the protocol again, otherwise the object is not thread safe,
                # this would cause safe failures later on
                self.exploration_step.protocol = []
        settings = self.exploration_step.options.__dict__
        setting_widget = ClassOptionsWidget(
            settings, self._docstring_dict, parent=self, add_close_button=True,
            allow_removal=False, suggestions_by_name=suggestions
        )
        setting_widget.exec_()
        self.exploration_step.options.__dict__.update(settings)
        self.changed_settings_signal.emit(self.exploration_step.options.__dict__)


def generate_gear_name_settings_suggestions(possible_gears: Optional[List[Gear]] = None) \
        -> Dict[str, Dict[str, Callable]]:
    """
    Utility function that allows us to generate an object that can be passed to DictOptionWidget,
    such that we can give the correct suggestion for the case of adding a new GearOptions entry.

    Returns
    -------
    Dict[str, Dict[str, Callable]]
        A dictionary with the names of gears and a callable that returns their settings
    """
    suggestions: Dict[str, Dict[str, Callable]] = {"gear_options": {}}
    if possible_gears is None:
        possible_gears = construct_all_possible_gears()
    suggestions["gear_options"] = {'gears': lambda: possible_gears}
    return suggestions


def construct_all_possible_gears() -> List[Gear]:
    """
    A utility function that constructs a list of all gears found by the ChemotonClassSearcher.

    Returns
    -------
    List[Gear]
        The list of found gears
    """
    gear_searcher = ChemotonClassSearcher(Gear)
    return [gear() for gear in gear_searcher.values()]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
from abc import abstractmethod
from typing import Dict, Optional, Any, List

from PySide2.QtGui import QPalette, QColor
from PySide2.QtWidgets import (
    QWidget,
    QLabel,
    QTextEdit,
)
from PySide2.QtCore import QObject


from scine_heron.containers.buttons import TextPushButton
from scine_heron.containers.layouts import HorizontalLayout, VerticalLayout
from scine_heron.containers.wrapped_label import WrappedLabel
from scine_heron.settings.docstring_parser import DocStringParser


class StartStopWidget(QWidget):
    """
    An abstract base class for widgets that hold a clas that can run some longer running process.
    And allows to start and stop this process by buttons, possibly interact with some settings,
    and display a status/progress by colors.

    Notes
    -----
    Does not inherit explicitly from ABC, because that would conflict with QObject.
    """

    # some predefined colors, possible to be moved to global style file
    green: str = "#5aa36b"
    orange: str = "#e8a01a"
    red: str = "#e8321a"

    def __init__(self, parent: Optional[QObject], *args, **kwargs):
        """
        Initialize with the underlying objects and construct the layout and buttons.
        """
        super().__init__(parent, *args, **kwargs)
        self._default_color = self.palette()
        self._layout = VerticalLayout()
        self._docstring_dict: Dict[str, str] = {}  # pylint: disable=unused-private-member
        self.button_settings: Optional[TextPushButton] = None
        self.button_delete: Optional[TextPushButton] = None
        self.button_copy_results: Optional[TextPushButton] = None
        self.filter_label: Optional[QLabel] = None
        self.filter: Optional[QTextEdit] = None
        self.init_arguments: List[Any] = [None]  # None for parent
        self.setLayout(self._layout)

    @abstractmethod
    def join(self, force_join: bool = False) -> Optional[Any]:
        """
        Join any forked process or threads

        Parameters
        ----------
        force_join : bool, optional
            If the process should also be joined, even if it never ran, by default False

        Returns
        -------
        Optional[Any]
           An optional returned object from the underlying process.
        """

    @abstractmethod
    def start_stop(self, start_all: bool = False, stop_all: bool = False) -> None:
        """
        Start or stop the underlying process depending on the current status

        Parameters
        ----------
        start_all : bool, optional
            Whether this method was triggered in a list of StartStopWidgets being started, by default False
        stop_all : bool, optional
            Whether this method was triggered in a list of StartStopWidgets being stopped, by default False
        """

    @abstractmethod
    def set_docstring_dict(self, doc_string_parser: DocStringParser) -> None:
        """
        Set the docstring information to the settings based on the given parser.

        Parameters
        ----------
        doc_string_parser : DocStringParser
            The parser that can extract the information from our objects.
        """

    @abstractmethod
    def stop_class_if_working(self) -> None:
        """
        Stop the underlying process if it is running
        """

    def _add_settings_and_delete_buttons(self):
        """
        Add two buttons to our layout that enable a settings access and a deletion of this widget.
        """
        self.button_settings = TextPushButton("Settings", self._show_settings)
        self.button_delete = TextPushButton("Delete")
        self._layout.add_widgets([self.button_settings, self.button_delete])

    def _show_settings(self) -> None:
        """
        Display the settings.
        """

    def _add_filter_at_layout(self, filter_description: str) -> None:
        """
        Add a textbox that describes the held filters with the given description.

        Parameters
        ----------
        filter_description : str
            The text to be displayed in the textbox.
        """

        self.filter_label = WrappedLabel("Filter(s):")
        self.filter = QTextEdit(filter_description)
        self.filter.setReadOnly(True)

        layout = HorizontalLayout([self.filter_label, self.filter])
        self._layout.addLayout(layout)

    def change_color(self, color: str = "") -> None:
        """
        Change the background color of this widget. If no color is given, the default color is used.

        Parameters
        ----------
        color : str, optional
            The color as a possible argument to the QColor constructor, by default ""
        """
        if color:
            palette = self.palette()
            palette.setColor(QPalette.Window, QColor(color))
            self.setPalette(palette)
            self.setAutoFillBackground(True)
        else:
            self.setPalette(self._default_color)
            self.setAutoFillBackground(False)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
import os
import signal
from argparse import ArgumentParser
from pathlib import Path
from typing import Optional

from PySide2.QtWidgets import QApplication, QMainWindow, QWidget
from qt_material import apply_stylesheet, get_theme
import scine_utilities as su

import scine_heron.config as config
from scine_heron.molecule.molecule_video import MainVideo
from scine_heron.resources import resource_path

from ._version import __version__  # noqa: F401


def find_main_window() -> Optional[QMainWindow]:
    # Global function to find the (open) QMainWindow in application
    app = QApplication.instance()
    for widget in app.topLevelWidgets():
        if isinstance(widget, QMainWindow):
            return widget
    return None


def get_core_tab(tab_identifier: str) -> Optional[QWidget]:
    main = find_main_window()
    if main is None:
        return None
    return main.get_tab(tab_identifier)


def view_trajectory_cli() -> None:
    parser = ArgumentParser(description='Enter filepath of trajectory')
    parser.add_argument(
        "file", help="load trajectory from file", type=Path, default=None,
    )
    parser.add_argument(
        "-m", "--mode", help="choose dark or light mode", type=str, default="dark"
    )
    args = parser.parse_args()
    theme_path = Path(os.path.join(resource_path(), f'theme_{args.mode}.xml'))
    view_trajectory_from_file(args.file, theme_path)


def view_trajectory_from_file(trajectory_path: Path, theme_path: Optional[Path] = None) -> None:
    p = trajectory_path
    ending = p.suffix
    if ending == ".bin":
        traj = su.io.read_trajectory(su.io.TrajectoryFormat.Binary, str(p.expanduser()))
    elif ending == ".xyz":
        traj = su.io.read_trajectory(su.io.TrajectoryFormat.Xyz, str(p.expanduser()))
    else:
        raise RuntimeError(f"Unsupported file type for trajectory {ending}")
    view_trajectory(traj, theme_path)


def view_trajectory(trajectory: su.MolecularTrajectory, theme_path: Optional[Path] = None) -> None:
    if theme_path is None:
        theme_path = Path(os.path.join(resource_path(), 'theme_dark.xml'))
    signal.signal(signal.SIGINT, signal.SIG_DFL)
    application = QApplication()
    css_path = theme_path.with_suffix(".xml")
    if os.path.exists(theme_path):
        apply_stylesheet(
            application,
            theme=str(theme_path),
            css_file=css_path,
            invert_secondary=True,
        )
        config.COLORS = get_theme(str(theme_path))
    window = MainVideo(trajectory)
    window.show()
    application.exec_()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
import sys
import traceback

from PySide2.QtCore import QObject, Signal, QRunnable, Slot

"""
Implemented according to:
https://www.pythonguis.com/tutorials/multithreading-pyside-applications-qthreadpool/
"""


class WorkerSignals(QObject):
    """
    Defines the signals available from a running worker thread.

    Supported Signals
    -----------------

    finished : Signal
        No data
    error : Signal
        tuple (exctype, value, traceback.format_exc() )
    result : Signal
        object data returned from processing, anything
    progress : Signal
        int indicating % progress
    running : Signal
        bool indicating if thread is running
    """

    finished = Signal()
    error = Signal(tuple)
    result = Signal(object)
    progress = Signal(int)
    running = Signal(tuple)


class Worker(QRunnable):
    """
    Worker thread

    Inherits from QRunnable to handler worker thread setup, signals and wrap-up.

    Parameters
    ----------
    callback : function
        The function callback to run on this worker thread. Supplied args and
        kwargs will be passed through to the runner.
    args
        Arguments to pass to the callback function
    kwargs
        Keywords to pass to the callback function
    """

    def __init__(self, fn, *args, **kwargs):
        super(Worker, self).__init__()

        # Store constructor arguments (re-used for processing)
        self.fn = fn
        self.args = args
        self.kwargs = kwargs
        self.signals = WorkerSignals()

        # Add the callback to our kwargs
        self.kwargs['progress_callback'] = self.signals.running

    @Slot()  # type: ignore
    def run(self):
        """
        Initialise the runner function with passed args, kwargs.
        """

        # Retrieve args/kwargs here; and fire processing using them
        try:
            result = self.fn(*self.args, **self.kwargs)
        except BaseException:
            traceback.print_exc()
            exctype, value = sys.exc_info()[:2]
            self.signals.error.emit((exctype, value, traceback.format_exc()))
        else:
            self.signals.result.emit(result)  # Return the result of the processing
        finally:
            self.signals.finished.emit()  # Done
